# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SVM.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='SVM.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\tSVM.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\x0e\n\x0cLinearKernel\"\x1a\n\tRBFKernel\x12\r\n\x05gamma\x18\x01 \x01(\x01\"6\n\nPolyKernel\x12\x0e\n\x06\x64\x65gree\x18\x01 \x01(\x05\x12\t\n\x01\x63\x18\x02 \x01(\x01\x12\r\n\x05gamma\x18\x03 \x01(\x01\")\n\rSigmoidKernel\x12\r\n\x05gamma\x18\x01 \x01(\x01\x12\t\n\x01\x63\x18\x02 \x01(\x01\"\xfa\x01\n\x06Kernel\x12:\n\x0clinearKernel\x18\x01 \x01(\x0b\x32\".CoreML.Specification.LinearKernelH\x00\x12\x34\n\trbfKernel\x18\x02 \x01(\x0b\x32\x1f.CoreML.Specification.RBFKernelH\x00\x12\x36\n\npolyKernel\x18\x03 \x01(\x0b\x32 .CoreML.Specification.PolyKernelH\x00\x12<\n\rsigmoidKernel\x18\x04 \x01(\x0b\x32#.CoreML.Specification.SigmoidKernelH\x00\x42\x08\n\x06kernel\"*\n\nSparseNode\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x01\"?\n\x0cSparseVector\x12/\n\x05nodes\x18\x01 \x03(\x0b\x32 .CoreML.Specification.SparseNode\"K\n\x14SparseSupportVectors\x12\x33\n\x07vectors\x18\x01 \x03(\x0b\x32\".CoreML.Specification.SparseVector\"\x1d\n\x0b\x44\x65nseVector\x12\x0e\n\x06values\x18\x01 \x03(\x01\"I\n\x13\x44\x65nseSupportVectors\x12\x32\n\x07vectors\x18\x01 \x03(\x0b\x32!.CoreML.Specification.DenseVector\"\x1d\n\x0c\x43oefficients\x12\r\n\x05\x61lpha\x18\x01 \x03(\x01\"\xb5\x02\n\x16SupportVectorRegressor\x12,\n\x06kernel\x18\x01 \x01(\x0b\x32\x1c.CoreML.Specification.Kernel\x12J\n\x14sparseSupportVectors\x18\x02 \x01(\x0b\x32*.CoreML.Specification.SparseSupportVectorsH\x00\x12H\n\x13\x64\x65nseSupportVectors\x18\x03 \x01(\x0b\x32).CoreML.Specification.DenseSupportVectorsH\x00\x12\x38\n\x0c\x63oefficients\x18\x04 \x01(\x0b\x32\".CoreML.Specification.Coefficients\x12\x0b\n\x03rho\x18\x05 \x01(\x01\x42\x10\n\x0esupportVectors\"\x8b\x04\n\x17SupportVectorClassifier\x12,\n\x06kernel\x18\x01 \x01(\x0b\x32\x1c.CoreML.Specification.Kernel\x12&\n\x1enumberOfSupportVectorsPerClass\x18\x02 \x03(\x05\x12J\n\x14sparseSupportVectors\x18\x03 \x01(\x0b\x32*.CoreML.Specification.SparseSupportVectorsH\x00\x12H\n\x13\x64\x65nseSupportVectors\x18\x04 \x01(\x0b\x32).CoreML.Specification.DenseSupportVectorsH\x00\x12\x38\n\x0c\x63oefficients\x18\x05 \x03(\x0b\x32\".CoreML.Specification.Coefficients\x12\x0b\n\x03rho\x18\x06 \x03(\x01\x12\r\n\x05probA\x18\x07 \x03(\x01\x12\r\n\x05probB\x18\x08 \x03(\x01\x12?\n\x11stringClassLabels\x18\x64 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x01\x12=\n\x10int64ClassLabels\x18\x65 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x01\x42\x10\n\x0esupportVectorsB\r\n\x0b\x43lassLabelsB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_LINEARKERNEL = _descriptor.Descriptor(
  name='LinearKernel',
  full_name='CoreML.Specification.LinearKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=57,
  serialized_end=71,
)


_RBFKERNEL = _descriptor.Descriptor(
  name='RBFKernel',
  full_name='CoreML.Specification.RBFKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.RBFKernel.gamma', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=99,
)


_POLYKERNEL = _descriptor.Descriptor(
  name='PolyKernel',
  full_name='CoreML.Specification.PolyKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='degree', full_name='CoreML.Specification.PolyKernel.degree', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='c', full_name='CoreML.Specification.PolyKernel.c', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.PolyKernel.gamma', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=155,
)


_SIGMOIDKERNEL = _descriptor.Descriptor(
  name='SigmoidKernel',
  full_name='CoreML.Specification.SigmoidKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.SigmoidKernel.gamma', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='c', full_name='CoreML.Specification.SigmoidKernel.c', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=198,
)


_KERNEL = _descriptor.Descriptor(
  name='Kernel',
  full_name='CoreML.Specification.Kernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='linearKernel', full_name='CoreML.Specification.Kernel.linearKernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rbfKernel', full_name='CoreML.Specification.Kernel.rbfKernel', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='polyKernel', full_name='CoreML.Specification.Kernel.polyKernel', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sigmoidKernel', full_name='CoreML.Specification.Kernel.sigmoidKernel', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kernel', full_name='CoreML.Specification.Kernel.kernel',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=201,
  serialized_end=451,
)


_SPARSENODE = _descriptor.Descriptor(
  name='SparseNode',
  full_name='CoreML.Specification.SparseNode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='CoreML.Specification.SparseNode.index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='CoreML.Specification.SparseNode.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=495,
)


_SPARSEVECTOR = _descriptor.Descriptor(
  name='SparseVector',
  full_name='CoreML.Specification.SparseVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nodes', full_name='CoreML.Specification.SparseVector.nodes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=497,
  serialized_end=560,
)


_SPARSESUPPORTVECTORS = _descriptor.Descriptor(
  name='SparseSupportVectors',
  full_name='CoreML.Specification.SparseSupportVectors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vectors', full_name='CoreML.Specification.SparseSupportVectors.vectors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=562,
  serialized_end=637,
)


_DENSEVECTOR = _descriptor.Descriptor(
  name='DenseVector',
  full_name='CoreML.Specification.DenseVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='CoreML.Specification.DenseVector.values', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=639,
  serialized_end=668,
)


_DENSESUPPORTVECTORS = _descriptor.Descriptor(
  name='DenseSupportVectors',
  full_name='CoreML.Specification.DenseSupportVectors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vectors', full_name='CoreML.Specification.DenseSupportVectors.vectors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=670,
  serialized_end=743,
)


_COEFFICIENTS = _descriptor.Descriptor(
  name='Coefficients',
  full_name='CoreML.Specification.Coefficients',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='alpha', full_name='CoreML.Specification.Coefficients.alpha', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=745,
  serialized_end=774,
)


_SUPPORTVECTORREGRESSOR = _descriptor.Descriptor(
  name='SupportVectorRegressor',
  full_name='CoreML.Specification.SupportVectorRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kernel', full_name='CoreML.Specification.SupportVectorRegressor.kernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sparseSupportVectors', full_name='CoreML.Specification.SupportVectorRegressor.sparseSupportVectors', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='denseSupportVectors', full_name='CoreML.Specification.SupportVectorRegressor.denseSupportVectors', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coefficients', full_name='CoreML.Specification.SupportVectorRegressor.coefficients', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rho', full_name='CoreML.Specification.SupportVectorRegressor.rho', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='supportVectors', full_name='CoreML.Specification.SupportVectorRegressor.supportVectors',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=777,
  serialized_end=1086,
)


_SUPPORTVECTORCLASSIFIER = _descriptor.Descriptor(
  name='SupportVectorClassifier',
  full_name='CoreML.Specification.SupportVectorClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kernel', full_name='CoreML.Specification.SupportVectorClassifier.kernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numberOfSupportVectorsPerClass', full_name='CoreML.Specification.SupportVectorClassifier.numberOfSupportVectorsPerClass', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sparseSupportVectors', full_name='CoreML.Specification.SupportVectorClassifier.sparseSupportVectors', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='denseSupportVectors', full_name='CoreML.Specification.SupportVectorClassifier.denseSupportVectors', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coefficients', full_name='CoreML.Specification.SupportVectorClassifier.coefficients', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rho', full_name='CoreML.Specification.SupportVectorClassifier.rho', index=5,
      number=6, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='probA', full_name='CoreML.Specification.SupportVectorClassifier.probA', index=6,
      number=7, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='probB', full_name='CoreML.Specification.SupportVectorClassifier.probB', index=7,
      number=8, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.stringClassLabels', index=8,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.int64ClassLabels', index=9,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='supportVectors', full_name='CoreML.Specification.SupportVectorClassifier.supportVectors',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.ClassLabels',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=1089,
  serialized_end=1612,
)

_KERNEL.fields_by_name['linearKernel'].message_type = _LINEARKERNEL
_KERNEL.fields_by_name['rbfKernel'].message_type = _RBFKERNEL
_KERNEL.fields_by_name['polyKernel'].message_type = _POLYKERNEL
_KERNEL.fields_by_name['sigmoidKernel'].message_type = _SIGMOIDKERNEL
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['linearKernel'])
_KERNEL.fields_by_name['linearKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['rbfKernel'])
_KERNEL.fields_by_name['rbfKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['polyKernel'])
_KERNEL.fields_by_name['polyKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['sigmoidKernel'])
_KERNEL.fields_by_name['sigmoidKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_SPARSEVECTOR.fields_by_name['nodes'].message_type = _SPARSENODE
_SPARSESUPPORTVECTORS.fields_by_name['vectors'].message_type = _SPARSEVECTOR
_DENSESUPPORTVECTORS.fields_by_name['vectors'].message_type = _DENSEVECTOR
_SUPPORTVECTORREGRESSOR.fields_by_name['kernel'].message_type = _KERNEL
_SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'].message_type = _SPARSESUPPORTVECTORS
_SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'].message_type = _DENSESUPPORTVECTORS
_SUPPORTVECTORREGRESSOR.fields_by_name['coefficients'].message_type = _COEFFICIENTS
_SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'])
_SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'].containing_oneof = _SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors']
_SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'])
_SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'].containing_oneof = _SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.fields_by_name['kernel'].message_type = _KERNEL
_SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'].message_type = _SPARSESUPPORTVECTORS
_SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'].message_type = _DENSESUPPORTVECTORS
_SUPPORTVECTORCLASSIFIER.fields_by_name['coefficients'].message_type = _COEFFICIENTS
_SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'].message_type = DataStructures__pb2._INT64VECTOR
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels']
DESCRIPTOR.message_types_by_name['LinearKernel'] = _LINEARKERNEL
DESCRIPTOR.message_types_by_name['RBFKernel'] = _RBFKERNEL
DESCRIPTOR.message_types_by_name['PolyKernel'] = _POLYKERNEL
DESCRIPTOR.message_types_by_name['SigmoidKernel'] = _SIGMOIDKERNEL
DESCRIPTOR.message_types_by_name['Kernel'] = _KERNEL
DESCRIPTOR.message_types_by_name['SparseNode'] = _SPARSENODE
DESCRIPTOR.message_types_by_name['SparseVector'] = _SPARSEVECTOR
DESCRIPTOR.message_types_by_name['SparseSupportVectors'] = _SPARSESUPPORTVECTORS
DESCRIPTOR.message_types_by_name['DenseVector'] = _DENSEVECTOR
DESCRIPTOR.message_types_by_name['DenseSupportVectors'] = _DENSESUPPORTVECTORS
DESCRIPTOR.message_types_by_name['Coefficients'] = _COEFFICIENTS
DESCRIPTOR.message_types_by_name['SupportVectorRegressor'] = _SUPPORTVECTORREGRESSOR
DESCRIPTOR.message_types_by_name['SupportVectorClassifier'] = _SUPPORTVECTORCLASSIFIER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LinearKernel = _reflection.GeneratedProtocolMessageType('LinearKernel', (_message.Message,), dict(
  DESCRIPTOR = _LINEARKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.LinearKernel)
  ))
_sym_db.RegisterMessage(LinearKernel)

RBFKernel = _reflection.GeneratedProtocolMessageType('RBFKernel', (_message.Message,), dict(
  DESCRIPTOR = _RBFKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.RBFKernel)
  ))
_sym_db.RegisterMessage(RBFKernel)

PolyKernel = _reflection.GeneratedProtocolMessageType('PolyKernel', (_message.Message,), dict(
  DESCRIPTOR = _POLYKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.PolyKernel)
  ))
_sym_db.RegisterMessage(PolyKernel)

SigmoidKernel = _reflection.GeneratedProtocolMessageType('SigmoidKernel', (_message.Message,), dict(
  DESCRIPTOR = _SIGMOIDKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SigmoidKernel)
  ))
_sym_db.RegisterMessage(SigmoidKernel)

Kernel = _reflection.GeneratedProtocolMessageType('Kernel', (_message.Message,), dict(
  DESCRIPTOR = _KERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Kernel)
  ))
_sym_db.RegisterMessage(Kernel)

SparseNode = _reflection.GeneratedProtocolMessageType('SparseNode', (_message.Message,), dict(
  DESCRIPTOR = _SPARSENODE,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseNode)
  ))
_sym_db.RegisterMessage(SparseNode)

SparseVector = _reflection.GeneratedProtocolMessageType('SparseVector', (_message.Message,), dict(
  DESCRIPTOR = _SPARSEVECTOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseVector)
  ))
_sym_db.RegisterMessage(SparseVector)

SparseSupportVectors = _reflection.GeneratedProtocolMessageType('SparseSupportVectors', (_message.Message,), dict(
  DESCRIPTOR = _SPARSESUPPORTVECTORS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseSupportVectors)
  ))
_sym_db.RegisterMessage(SparseSupportVectors)

DenseVector = _reflection.GeneratedProtocolMessageType('DenseVector', (_message.Message,), dict(
  DESCRIPTOR = _DENSEVECTOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DenseVector)
  ))
_sym_db.RegisterMessage(DenseVector)

DenseSupportVectors = _reflection.GeneratedProtocolMessageType('DenseSupportVectors', (_message.Message,), dict(
  DESCRIPTOR = _DENSESUPPORTVECTORS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DenseSupportVectors)
  ))
_sym_db.RegisterMessage(DenseSupportVectors)

Coefficients = _reflection.GeneratedProtocolMessageType('Coefficients', (_message.Message,), dict(
  DESCRIPTOR = _COEFFICIENTS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Coefficients)
  ))
_sym_db.RegisterMessage(Coefficients)

SupportVectorRegressor = _reflection.GeneratedProtocolMessageType('SupportVectorRegressor', (_message.Message,), dict(
  DESCRIPTOR = _SUPPORTVECTORREGRESSOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SupportVectorRegressor)
  ))
_sym_db.RegisterMessage(SupportVectorRegressor)

SupportVectorClassifier = _reflection.GeneratedProtocolMessageType('SupportVectorClassifier', (_message.Message,), dict(
  DESCRIPTOR = _SUPPORTVECTORCLASSIFIER,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SupportVectorClassifier)
  ))
_sym_db.RegisterMessage(SupportVectorClassifier)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

"""
Functions related to graph visualization of mlmodels
"""

import ast as _ast
import json as _json
import os as _os
import numpy as _np
from ._infer_shapes_nn_mlmodel import infer_shapes as _infer_shapes
from coremltools.proto import NeuralNetwork_pb2 as _NeuralNetwork_pb2


def _calculate_edges(cy_nodes, cy_edges, shape_dict=None):
    """

    Parameters
    ----------
    cy_nodes: list of nodes for graph
    cy_edges: list of edges to be updated for graph
    shape_dict: shape_dict required for inferring shape information

    Returns
    -------

    cy_nodes: list of nodes for graph
    cy_edges: list of edges to be updated for graph

    """
    node_len = len(cy_nodes)

    for upper_index in range(0, node_len):
        for lower_index in range(upper_index + 1, node_len):

            if (
                "outputs" in cy_nodes[upper_index]["data"]["info"].keys()
                and "inputs" in cy_nodes[upper_index]["data"]["info"].keys()
                and "outputs" in cy_nodes[lower_index]["data"]["info"].keys()
                and "inputs" in cy_nodes[lower_index]["data"]["info"].keys()
            ):
                outputs = _ast.literal_eval(
                    cy_nodes[upper_index]["data"]["info"]["outputs"]
                )
                inputs = _ast.literal_eval(
                    cy_nodes[lower_index]["data"]["info"]["inputs"]
                )
                for output in outputs:
                    if output in inputs:
                        if shape_dict is None or output not in shape_dict.keys():
                            label = None
                        else:
                            label = str(shape_dict[output])

                        cy_edges.append(
                            {
                                "data": {
                                    "id": "{}.{}.{}".format(
                                        output,
                                        cy_nodes[upper_index]["data"]["id"],
                                        cy_nodes[lower_index]["data"]["id"],
                                    ),
                                    "source": cy_nodes[upper_index]["data"]["id"],
                                    "target": cy_nodes[lower_index]["data"]["id"],
                                    "label": label,
                                    "shape": label,
                                }
                            }
                        )

    return cy_nodes, cy_edges


def _layer_specific_info(layer):
    """

    Parameters
    ----------
    layer : Can be one of : 'activation', 'add', 'average', 'batchnorm',
    'biDirectionalLSTM', 'bias', 'concat', 'convolution', 'crop', 'dot',
    'embedding', 'flatten', 'gru', 'innerProduct', 'input', 'l2normalize',
    'loadConstant', 'lrn', 'max', 'min', 'multiply', 'mvn', 'name', 'output',
    'padding', permute', 'pooling', 'reduce', 'reorganizeData', 'reshape',
    'scale', 'sequenceRepeat', 'simpleRecurrent', 'slice', 'softmax', 'split',
     'unary', 'uniDirectionalLSTM', 'upsample'

    Returns
    -------
    info : info specific to layer type

    """
    if layer.WhichOneof("layer") == "convolution":
        info = {
            "type": layer.WhichOneof("layer"),
            "outputChannels": _json.dumps(str(layer.convolution.outputChannels)),
            "kernelChannels": _json.dumps(str(layer.convolution.kernelChannels)),
            "groups": _json.dumps(str(layer.convolution.nGroups)),
            "kernelSize": _json.dumps(str(layer.convolution.kernelSize)),
            "stride": _json.dumps(str(layer.convolution.stride)),
            "dilationFactor": _json.dumps(str(layer.convolution.dilationFactor)),
            "isDeconvolution": _json.dumps(str(layer.convolution.isDeconvolution)),
            "paddingType": _json.dumps(
                layer.convolution.WhichOneof("ConvolutionPaddingType")
            ),
            "desc": "A layer that performs spatial convolution",
        }
        if _json.dumps(layer.convolution.isDeconvolution) == "true":
            info["type"] = "deconvolution"
            info["desc"] = "A layer that performs spatial deconvolution"

    elif layer.WhichOneof("layer") == "activation":
        params = layer.activation
        act_type = params.WhichOneof("NonlinearityType")
        info = {
            "type": layer.WhichOneof("layer"),
            "activationType": act_type,
            "desc": "Applies specified type of activation function to input.",
        }
        if act_type == "linear":
            info["alpha"] = _json.dumps(str(params.linear.alpha))
            info["beta"] = _json.dumps(str(params.linear.beta))
        if act_type == "leakyReLU":
            info["alpha"] = _json.dumps(str(params.leakyReLU.alpha))
        if act_type == "thresholdedReLU":
            info["alpha"] = _json.dumps(str(params.thresholdedReLU.alpha))
        if act_type == "scaledTanh":
            info["alpha"] = _json.dumps(str(params.scaledTanh.alpha))
            info["beta"] = _json.dumps(str(params.scaledTanh.beta))
        if act_type == "sigmoidHard":
            info["alpha"] = _json.dumps(str(params.sigmoidHard.alpha))
            info["beta"] = _json.dumps(str(params.sigmoidHard.beta))
        if act_type == "ELU":
            info["alpha"] = _json.dumps(str(params.ELU.alpha))

    elif layer.WhichOneof("layer") == "pooling":
        params = layer.pooling
        paddingType = params.WhichOneof("PoolingPaddingType")
        info = {
            "type": layer.WhichOneof("layer"),
            "desc": "Spatial Pooling layer to reduce dimensions of input using the "
            "specified kernel size and type.",
        }
        if params.globalPooling:
            info["globalPooling"] = "True"
            info["poolingType"] = "global pooling"
        else:
            info["poolingType"] = _json.dumps(
                _NeuralNetwork_pb2.PoolingLayerParams.PoolingType.Name(params.type)
            )
            info["stride"] = _json.dumps(str(params.stride))
            info["kernelSize"] = _json.dumps(str(params.kernelSize))
            info["paddingType"] = _json.dumps(paddingType)

    elif layer.WhichOneof("layer") == "add":
        info = {
            "type": layer.WhichOneof("layer"),
            "alpha": _json.dumps(str(layer.add.alpha)),
            "desc": "A layer that performs elementwise addition.",
        }
    elif layer.WhichOneof("layer") == "batchnorm":
        info = {
            "type": layer.WhichOneof("layer"),
            "channels": _json.dumps(str(layer.batchnorm.channels)),
            "computeMeanVar": _json.dumps(str(layer.batchnorm.computeMeanVar)),
            "instanceNormalization": _json.dumps(
                str(layer.batchnorm.instanceNormalization)
            ),
            "desc": "A layer that performs batch normalization, \n"
            "which is performed along the channel axis, \n"
            "and repeated along the other axes, if present.",
        }
    elif layer.WhichOneof("layer") == "biDirectionalLSTM":
        forward_activations = ""
        for activation in layer.biDirectionalLSTM.activationsForwardLSTM:
            forward_activations += str(activation)[:-5] + ", "
        backward_activations = ""
        for activation in layer.biDirectionalLSTM.activationsBackwardLSTM:
            backward_activations += str(activation)[:-5] + ", "
        info = {
            "type": layer.WhichOneof("layer"),
            "inputVectorSize": _json.dumps(
                str(layer.biDirectionalLSTM.inputVectorSize)
            ),
            "outputVectorSize": _json.dumps(
                str(layer.biDirectionalLSTM.outputVectorSize)
            ),
            "forward_activations": _json.dumps(forward_activations),
            "backward_activations": _json.dumps(backward_activations),
            "lstm_params": _json.dumps(str(layer.biDirectionalLSTM.params)),
            "desc": "Bidirectional long short-term memory (LSTM) layer\n"
            "The first LSTM operates on the input sequence in the forward direction.\n"
            "The second LSTM operates on the input sequence in the reverse direction.",
        }
    elif layer.WhichOneof("layer") == "uniDirectionalLSTM":
        activations = ""
        for activation in layer.uniDirectionalLSTM.activations:
            activations += str(activation)[:-5] + ", "
        info = {
            "type": layer.WhichOneof("layer"),
            "inputVectorSize": _json.dumps(
                str(layer.uniDirectionalLSTM.inputVectorSize)
            ),
            "outputVectorSize": _json.dumps(
                str(layer.uniDirectionalLSTM.outputVectorSize)
            ),
            "activations": _json.dumps(activations),
            "lstm_params": _json.dumps(str(layer.uniDirectionalLSTM.params)),
            "reverse_input": _json.dumps(str(layer.uniDirectionalLSTM.reverseInput)),
            "desc": "A unidirectional long short-term memory (LSTM) layer.",
        }
    elif layer.WhichOneof("layer") == "gru":
        activations = ""
        for activation in layer.gru.activations:
            activations += str(activation)[:-5] + ", "
        info = {
            "type": layer.WhichOneof("layer"),
            "inputVectorSize": _json.dumps(str(layer.gru.inputVectorSize)),
            "outputVectorSize": _json.dumps(str(layer.gru.outputVectorSize)),
            "activations": _json.dumps(activations),
            "hasBiasVectors": _json.dumps(str(layer.gru.hasBiasVectors)),
            "reverseInput": _json.dumps(str(layer.gru.reverseInput)),
            "sequenceOutput": _json.dumps(str(layer.gru.sequenceOutput)),
            "desc": "Gated-Recurrent Unit (GRU) Layer.\n",
        }
    elif layer.WhichOneof("layer") == "simpleRecurrent":
        info = {
            "type": layer.WhichOneof("layer"),
            "inputVectorSize": _json.dumps(str(layer.simpleRecurrent.inputVectorSize)),
            "outputVectorSize": _json.dumps(
                str(layer.simpleRecurrent.outputVectorSize)
            ),
            "activation": _json.dumps(str(layer.simpleRecurrent.activation)),
            "hasBiasVector": _json.dumps(str(layer.simpleRecurrent.hasBiasVector)),
            "reverseInput": _json.dumps(str(layer.simpleRecurrent.reverseInput)),
            "sequenceOutput": _json.dumps(str(layer.simpleRecurrent.sequenceOutput)),
            "desc": "A simple recurrent layer.",
        }
    elif layer.WhichOneof("layer") == "bias":
        info = {
            "type": layer.WhichOneof("layer"),
            "shape": _json.dumps(str(layer.bias.shape)),
            "desc": "A layer that performs elementwise addition of a bias,\n"
            "which is broadcasted to match the input shape.",
        }
    elif layer.WhichOneof("layer") == "concat":
        info = {
            "type": layer.WhichOneof("layer"),
            "sequenceConcat": _json.dumps(str(layer.concat.sequenceConcat)),
            "desc": "A layer that concatenates along the channel axis (default) or sequence axis.",
        }
    elif layer.WhichOneof("layer") == "crop":
        info = {
            "type": layer.WhichOneof("layer"),
            "cropAmounts": _json.dumps(str(layer.crop.cropAmounts)),
            "offset": _json.dumps(str(layer.crop.offset)),
            "desc": "A layer that crops the spatial dimensions of an input.\n"
            "If two inputs are provided, the shape of the second "
            "input is used as the reference shape.",
        }
    elif layer.WhichOneof("layer") == "dot":
        info = {
            "type": layer.WhichOneof("layer"),
            "cosineSimilarity": _json.dumps(str(layer.dot.cosineSimilarity)),
            "desc": "If true, inputs are normalized first, "
            "thereby computing the cosine similarity.",
        }
    elif layer.WhichOneof("layer") == "embedding":
        info = {
            "type": layer.WhichOneof("layer"),
            "inputDim": _json.dumps(str(layer.embedding.inputDim)),
            "outputChannels": _json.dumps(str(layer.embedding.outputChannels)),
            "hasBias": _json.dumps(str(layer.embedding.inputDim)),
            "desc": "A layer that performs a matrix lookup and optionally adds a bias.",
        }
    elif layer.WhichOneof("layer") == "flatten":
        info = {
            "type": layer.WhichOneof("layer"),
            "mode": _json.dumps(
                _NeuralNetwork_pb2.FlattenLayerParams.FlattenOrder.Name(
                    layer.flatten.mode
                )
            ),
            "desc": "A layer that flattens the input.",
        }
    elif layer.WhichOneof("layer") == "innerProduct":
        info = {
            "type": layer.WhichOneof("layer"),
            "inputChannels": _json.dumps(str(layer.innerProduct.inputChannels)),
            "outputChannels": _json.dumps(str(layer.innerProduct.outputChannels)),
            "hasBias": _json.dumps(str(layer.innerProduct.hasBias)),
            "desc": "A layer that performs a matrix vector product.\n"
            "This is equivalent to a fully-connected, or dense layer.",
        }
    elif layer.WhichOneof("layer") == "l2normalize":
        info = {
            "type": layer.WhichOneof("layer"),
            "epsilon": _json.dumps(str(layer.l2normalize.epsilon)),
            "desc": "A layer that performs L2 normalization, i.e. divides by the \n"
            "the square root of the sum of squares of all elements of input.",
        }
    elif layer.WhichOneof("layer") == "loadConstant":
        info = {
            "type": layer.WhichOneof("layer"),
            "shape": _json.dumps(str(layer.loadConstant.shape)),
            "desc": "The shape of the constant to be loaded",
        }
    elif layer.WhichOneof("layer") == "lrn":
        info = {
            "type": layer.WhichOneof("layer"),
            "alpha": _json.dumps(str(layer.lrn.alpha)),
            "beta": _json.dumps(str(layer.lrn.beta)),
            "localSize": _json.dumps(str(layer.lrn.localSize)),
            "k": _json.dumps(str(layer.lrn.k)),
            "desc": "A layer that performs local response normalization (LRN).",
        }
    elif layer.WhichOneof("layer") == "multiply":
        info = {
            "type": layer.WhichOneof("layer"),
            "alpha": _json.dumps(str(layer.multiply.alpha)),
            "desc": "A layer that performs elementwise multiplication.",
        }
    elif layer.WhichOneof("layer") == "mvn":
        info = {
            "type": layer.WhichOneof("layer"),
            "acrossChannels": _json.dumps(str(layer.mvn.acrossChannels)),
            "normalizeVariance": _json.dumps(str(layer.mvn.normalizeVariance)),
            "epsilon": _json.dumps(str(layer.mvn.epsilon)),
            "desc": "A layer that performs mean variance normalization.",
        }
    elif layer.WhichOneof("layer") == "padding":
        info = {
            "type": layer.WhichOneof("layer"),
            "paddingAmounts": _json.dumps(str(layer.padding.paddingAmounts)),
            "paddingType": _json.dumps(str(layer.padding.WhichOneof("PaddingType"))),
            "desc": "Fill a constant value in the padded region.",
        }
    elif layer.WhichOneof("layer") == "permute":
        info = {
            "type": layer.WhichOneof("layer"),
            "axis": _json.dumps(str(layer.permute.axis)),
            "desc": "A layer that rearranges the dimensions and data of an input.",
        }
    elif layer.WhichOneof("layer") == "reduce":
        params = layer.reduce
        info = {
            "type": layer.WhichOneof("layer"),
            "mode": _json.dumps(str(params.mode)),
            "epsilon": _json.dumps(str(params.epsilon)),
            "axis": _json.dumps(
                _NeuralNetwork_pb2.ReduceLayerParams.ReduceAxis.Name(params.axis)
            ),
            "desc": "A layer that reduces the input using a specified operation.",
        }
    elif layer.WhichOneof("layer") == "reorganizeData":
        info = {
            "type": layer.WhichOneof("layer"),
            "mode": _json.dumps(
                _NeuralNetwork_pb2.ReorganizeDataLayerParams.ReorganizationType.Name(
                    layer.reorganizeData.mode
                )
            ),
            "blockSize": _json.dumps(str(layer.reorganizeData.blockSize)),
            "desc": "A layer that reorganizes data in the input in: \n"
            "1. SPACE_TO_DEPTH\n"
            "2. DEPTH_TO_SPACE",
        }
    elif layer.WhichOneof("layer") == "reshape":
        info = {
            "type": layer.WhichOneof("layer"),
            "mode": _json.dumps(
                _NeuralNetwork_pb2.ReshapeLayerParams.ReshapeOrder.Name(
                    layer.reshape.mode
                )
            ),
            "targetShape": _json.dumps(str(layer.reshape.targetShape)),
            "desc": "A layer that recasts the input into a new shape.",
        }
    elif layer.WhichOneof("layer") == "scale":
        info = {
            "type": layer.WhichOneof("layer"),
            "shapeScale": _json.dumps(str(layer.scale.shapeScale)),
            "hasBias": _json.dumps(str(layer.scale.hasBias)),
            "shapeBias": _json.dumps(str(layer.scale.shapeBias)),
            "desc": "A layer that performs elmentwise multiplication by a scale factor\n"
            "and optionally adds a bias;",
        }
    elif layer.WhichOneof("layer") == "sequenceRepeat":
        info = {
            "type": layer.WhichOneof("layer"),
            "nRepetitions": _json.dumps(str(layer.sequenceRepeat.nRepetitions)),
            "desc": "A layer that repeats a sequence.",
        }
    elif layer.WhichOneof("layer") == "slice":
        info = {
            "type": layer.WhichOneof("layer"),
            "startIndex": _json.dumps(str(layer.slice.startIndex)),
            "endIndex": _json.dumps(str(layer.slice.endIndex)),
            "stride": _json.dumps(str(layer.slice.stride)),
            "axis": _json.dumps(
                _NeuralNetwork_pb2.SliceLayerParams.SliceAxis.Name(layer.slice.axis)
            ),
            "desc": "A layer that slices the input data along a given axis.",
        }
    elif layer.WhichOneof("layer") == "split":
        info = {
            "type": layer.WhichOneof("layer"),
            "nOutputs": _json.dumps(str(layer.split.nOutputs)),
            "desc": "A layer that uniformly splits across the channel dimension\n"
            "to produce a specified number of outputs.",
        }
    elif layer.WhichOneof("layer") == "unary":
        info = {
            "type": layer.WhichOneof("layer"),
            "unary_type": _json.dumps(
                _NeuralNetwork_pb2.UnaryFunctionLayerParams.Operation.Name(
                    layer.unary.type
                )
            ),
            "alpha": _json.dumps(str(layer.unary.alpha)),
            "epsilon": _json.dumps(str(layer.unary.epsilon)),
            "shift": _json.dumps(str(layer.unary.shift)),
            "scale": _json.dumps(str(layer.unary.scale)),
            "desc": "A layer that applies a unary function.",
        }
    elif layer.WhichOneof("layer") == "upsample":
        info = {
            "type": layer.WhichOneof("layer"),
            "scalingFactor": _json.dumps(str(layer.upsample.scalingFactor)),
            "mode": _json.dumps(
                _NeuralNetwork_pb2.UpsampleLayerParams.InterpolationMode.Name(
                    layer.upsample.mode
                )
            ),
            "desc": "A layer that scales up spatial dimensions.\n"
            "It supports two modes: "
            "nearest neighbour (default) and bilinear.",
        }
    elif layer.WhichOneof("layer") == "max":
        info = {
            "type": layer.WhichOneof("layer"),
            "desc": "A layer that computes the elementwise maximum " "over the inputs.",
        }
    elif layer.WhichOneof("layer") == "min":
        info = {
            "type": layer.WhichOneof("layer"),
            "desc": "A layer that computes the elementwise minimum " "over the inputs.",
        }
    elif layer.WhichOneof("layer") == "average":
        info = {
            "type": layer.WhichOneof("layer"),
            "desc": "A layer that computes the elementwise average " "of the inputs.",
        }
    elif layer.WhichOneof("layer") == "softmax":
        info = {
            "type": layer.WhichOneof("layer"),
            "desc": "A layer that performs softmax normalization.\n"
            "Normalization is done along the channel axis.",
        }
    elif layer.WhichOneof("layer") == "custom":
        info = {
            "type": layer.WhichOneof("layer"),
            "className": layer.custom.className,
            "desc": "A custom layer",
        }
        if layer.custom.parameters != {}:
            for key in layer.custom.parameters.keys():
                value = _get_custom_layer_value(layer.custom.parameters[key])
                info[key] = value
        if layer.custom.description:
            info["desc"] = layer.custom.description

    else:
        info = {"type": layer.WhichOneof("layer")}

    info["inputs"] = str(layer.input)
    info["outputs"] = str(layer.output)

    return info


def _get_custom_layer_value(parameter):
    if "intValue" in str(parameter):
        return str(parameter.intValue)
    elif "doubleValue" in str(parameter):
        return str(parameter.doubleValue)
    elif "boolValue" in str(parameter):
        return str(parameter.boolValue)
    elif "longValue" in str(parameter):
        return str(parameter.longValue)
    elif "stringValue" in str(parameter):
        return str(parameter.stringValue)


def _pipeline_component_info(model, info):
    """

    Parameters
    ----------
    model : pipeline model
    info : info dict to dump model related info into

    model can be one of 'arrayFeatureExtractor', 'categoricalMapping',
    'dictVectorizer', 'featureVectorizer', 'glmClassifier', 'glmRegressor',
    'identity', 'imputer', 'neuralNetwork', 'neuralNetworkClassifier',
    'neuralNetworkRegressor', 'normalizer', 'oneHotEncoder', 'scaler',
    'supportVectorClassifier', 'supportVectorRegressor',
    'treeEnsembleClassifier', 'treeEnsembleRegressor'

    Returns
    -------
    info : info dict with required info for model

    """
    model_type = model.WhichOneof("Type")
    if model_type == "arrayFeatureExtractor":
        info["desc"] = (
            "Given an index, extracts the value at "
            "that index from its array input.\n"
            "Indexes are zero-based."
        )
    elif model_type == "categoricalMapping":
        info["mappingType"] = _json.dumps(
            str(model.categoricalMapping.WhichOneof("MappingType"))
        )
        info["valueOnUnknown"] = _json.dumps(
            str(model.categoricalMapping.WhichOneof("ValueOnUnknown"))
        )
        info["desc"] = (
            "This allows conversion from integers "
            "to strings, or from strings to integers."
        )
    elif model_type == "dictVectorizer":
        info["map"] = _json.dumps(str(model.dictVectorizer.WhichOneof("Map")))
        info["desc"] = (
            "Uses an index mapping to convert a dictionary "
            "to an array.\n The output array will be equal in "
            "length to the index mapping vector parameter.\n"
            "All keys in the input dictionary must be present in "
            "the index mapping vector."
        )
    elif model_type == "featureVectorizer":
        info["inputList"] = _json.dumps(str(model.featureVectorizer.inputList))
        info["desc"] = (
            "A FeatureVectorizer puts one or more features into a single"
            " array.\n The ordering of features in the output array is "
            "determined by inputList."
        )
    elif model_type == "glmClassifier":
        info["offset"] = _json.dumps(str(model.glmClassifier.offset))
        info["postEvaluationTransform"] = _json.dumps(
            str(model.glmClassifier.postEvaluationTransform)
        )
        info["classEncoding"] = _json.dumps(str(model.glmClassifier.classEncoding))
        info["classLabels"] = _json.dumps(
            str(model.glmClassifier.WhichOneof("ClassLabels"))
        )
        info["desc"] = "A generalized linear model classifier."
    elif model_type == "glmRegressor":
        info["offset"] = _json.dumps(str(model.glmRegressor.offset))
        info["postEvaluationTransform"] = _json.dumps(
            str(model.glmRegressor.postEvaluationTransform)
        )
        info["desc"] = "A generalized linear model regressor."
    elif model_type == "imputer":
        info["ImputedValue"] = _json.dumps(
            str(model.imputer.WhichOneof("ImputedValue"))
        )
        info["desc"] = (
            "A transformer that replaces missing values with a "
            "default value,\n such as a statistically-derived "
            "value.\nIf ``ReplaceValue`` is set, then missing "
            "values of that type are\n replaced with the "
            "corresponding value."
        )
    elif model_type == "normalizer":
        info["normType"] = _json.dumps(str(model.normalizer.normType))
        info["desc"] = (
            "A normalization preprocessor.There are three normalization modes\n"
            "1. Max\n"
            "2. L1\n"
            "3. L2"
        )
    elif model_type == "oneHotEncoder":
        info["CategoryType"] = _json.dumps(
            str(model.oneHotEncoder.WhichOneof("CategoryType"))
        )
        info["outputSparse"] = _json.dumps(str(model.oneHotEncoder.outputSparse))
        info["handleUnknown"] = _json.dumps(str(model.oneHotEncoder.handleUnknown))
        info["desc"] = (
            "Transforms a categorical feature into an array. The array will be all\n"
            "zeros expect a single entry of one.\n"
            "Each categorical value will map to an index, this mapping is given by\n"
            "either the ``stringCategories`` parameter or the ``int64Categories``\n"
            "parameter."
        )
    elif model_type == "scaler":
        info["shiftValue"] = _json.dumps(str(model.scaler.shiftValue))
        info["scaleValue"] = _json.dumps(str(model.scaler.scaleValue))
        info["desc"] = (
            "A scaling operation.\n" r"f(x) = scaleValue \cdot (x + shiftValue)"
        )
    elif model_type == "supportVectorClassifier":
        info["kernel"] = _json.dumps(str(model.supportVectorClassifier.kernel))
        info["numberOfSupportVectorsPerClass"] = _json.dumps(
            str(model.supportVectorClassifier.numberOfSupportVectorsPerClass)
        )
        info["rho"] = _json.dumps(str(model.supportVectorClassifier.rho))
        info["probA"] = _json.dumps(str(model.supportVectorClassifier.probA))
        info["probB"] = _json.dumps(str(model.supportVectorClassifier.probB))
        info["ClassLabels"] = _json.dumps(
            str(model.supportVectorClassifier.WhichOneof("ClassLabels"))
        )
        info["desc"] = (
            "Support Vector Machine Classifier with one of "
            "Linear, RBF, Polynomial or Sigmoid "
            "kernels available"
        )
    elif model_type == "supportVectorRegressor":
        info["kernel"] = _json.dumps(str(model.supportVectorRegressor.kernel))
        info["numberOfSupportVectorsPerClass"] = _json.dumps(
            str(model.supportVectorRegressor.numberOfSupportVectorsPerClass)
        )
        info["rho"] = _json.dumps(str(model.supportVectorRegressor.rho))
        info["desc"] = (
            "Support Vector Machine Regressor with one of "
            "Linear, RBF, Polynomial or Sigmoid kernels available"
        )
    elif model_type == "treeEnsembleClassifier":
        info["treeEnsemble"] = _json.dumps(
            str(model.treeEnsembleClassifier.treeEnsemble)
        )
        info["postEvaluationTransform"] = _json.dumps(
            str(model.treeEnsembleClassifier.postEvaluationTransform)
        )
        info["ClassLabels"] = _json.dumps(
            str(model.treeEnsembleClassifier.WhichOneof("ClassLabels"))
        )
        info["desc"] = (
            "Each tree is a collection of nodes, each of which is identified "
            "by a unique identifier.\nEach node is either a branch or a leaf node."
            " A branch node evaluates a value according to a behavior;\n"
            "A tree must have exactly one root node, which has no parent node."
        )
    elif model_type == "treeEnsembleRegressor":
        info["treeEnsemble"] = _json.dumps(
            str(model.treeEnsembleRegressor.treeEnsemble)
        )
        info["postEvaluationTransform"] = _json.dumps(
            str(model.treeEnsembleRegressor.postEvaluationTransform)
        )
        info["desc"] = (
            "Each tree is a collection of nodes, each of which is identified"
            " by a unique identifier.\nEach node is either a branch or a leaf"
            " node. A branch node evaluates a value according to a behavior;\n"
            "A tree must have exactly one root node, which has no parent node."
        )
    return info


def _neural_network_node_info(nn_spec, cy_nodes, child=False, parent=None):
    """

    Parameters
    ----------
    nn_spec : Neural Network spec of mlmodel
    cy_nodes: list of nodes to update with nn layers
    child: If child of a parent pipeline component
    parent : Parent node of the Neural Network spec

    Returns
    -------

    cy_nodes: Updated with layer specific information

    """
    layers = nn_spec.layers
    for layer in layers:
        info = _layer_specific_info(layer)
        if child:
            info["name"] = layer.name
            cy_nodes.append(
                {
                    "data": {
                        "id": layer.name,
                        "name": info["type"],
                        "info": info,
                        "parent": parent,
                    },
                    "classes": info["type"],
                }
            )
        else:
            info["name"] = layer.name
            cy_nodes.append(
                {
                    "data": {"id": layer.name, "name": info["type"], "info": info},
                    "classes": info["type"],
                }
            )

    return cy_nodes


def _neural_network_nodes_and_edges(
    nn_spec, cy_nodes, cy_edges, spec_outputs, input_spec, input_shape_dict=None
):
    """

    Parameters
    ----------
    nn_spec : Neural Network Spec
    cy_nodes : list to add nn nodes to
    cy_edges : list to add edges for nn nodes to
    spec_outputs : outputs of nn spec
    input_spec : input spec of Neural Network

    Returns
    -------

    cy_data : concatenated list of updated cy_nodes and cy_edges

    """
    cy_nodes = _neural_network_node_info(nn_spec, cy_nodes)
    cy_nodes.append(
        {
            "data": {
                "id": "output_node",
                "name": "",
                "info": {"type": "output node"},
                "classes": "output",
            }
        }
    )

    for model_output, output_type in spec_outputs:
        cy_nodes.append(
            {
                "data": {
                    "id": str(model_output),
                    "name": str(model_output),
                    "info": {
                        "type": "\n".join(str(output_type).split("\n")),
                        "inputs": str([model_output]),
                        "outputs": str([]),
                    },
                    "parent": "output_node",
                },
                "classes": "output",
            }
        )

    shape_dict = _infer_shapes(nn_spec, input_spec, input_shape_dict=input_shape_dict)
    cy_nodes, cy_edges = _calculate_edges(cy_nodes, cy_edges, shape_dict)

    cy_data = cy_nodes + cy_edges
    return cy_data


def _pipeline_nodes_and_edges(cy_nodes, cy_edges, pipeline_spec, spec_outputs):
    """

    Parameters
    ----------
    cy_nodes : list to add nn nodes to
    cy_edges : list to add edges for nn nodes to
    pipeline_spec: Spec of pipeline mlmodel
    spec_outputs: spec outputs of pipeline mlmodel

    Returns
    -------

    cy_data : concatenated list of updated cy_nodes and cy_edges

    """
    i = 1
    nn_model_types = [
        "neuralNetwork",
        "neuralNetworkClassifier",
        "neuralNetworkRegressor",
    ]
    models = pipeline_spec.models
    shape_dict = None
    for model in models:
        sub_model_type = model.WhichOneof("Type")
        if not sub_model_type:
            sub_model_type = "input"
        info = {}
        input_names = []
        output_names = []
        info["Pipeline Component"] = sub_model_type.upper()
        for model_input in model.description.input:
            input_names.append(model_input.name)
            info["inputs"] = str(input_names)

        for model_output in model.description.output:
            output_names.append(model_output.name)
            info["outputs"] = str(output_names)

        info = _pipeline_component_info(model, info)

        if sub_model_type in nn_model_types:
            cy_nodes.append(
                {
                    "data": {
                        "id": "{}_{}".format(sub_model_type, i),
                        "name": sub_model_type,
                        "info": info,
                    },
                    "classes": "parent",
                }
            )
            if sub_model_type == "neuralNetwork":
                nn_spec = model.neuralNetwork
            elif sub_model_type == "neuralNetworkClassifier":
                nn_spec = model.neuralNetworkClassifier
            elif sub_model_type == "neuralNetworkRegressor":
                nn_spec = model.neuralNetworkRegressor
            cy_nodes = _neural_network_node_info(
                nn_spec, cy_nodes, child=True, parent="{}_{}".format(sub_model_type, i)
            )
            shape_dict = _infer_shapes(nn_spec, model.description.input)
        else:
            cy_nodes.append(
                {
                    "data": {
                        "id": "{}_{}".format(sub_model_type, i),
                        "name": sub_model_type,
                        "info": info,
                    },
                    "classes": sub_model_type,
                }
            )
        i += 1

    cy_nodes.append(
        {
            "data": {
                "id": "output_node",
                "name": "",
                "info": {"type": "output node"},
                "classes": "output",
            }
        }
    )

    for model_output, output_type in spec_outputs:
        cy_nodes.append(
            {
                "data": {
                    "id": str(model_output),
                    "name": str(model_output),
                    "info": {
                        "type": "\n".join(str(output_type).split("\n")),
                        "inputs": str([model_output]),
                        "outputs": str([]),
                    },
                    "parent": "output_node",
                },
                "classes": "output",
            }
        )

    cy_nodes, cy_edges = _calculate_edges(cy_nodes, cy_edges, shape_dict)

    cy_data = cy_nodes + cy_edges
    return cy_data


def _start_server(port, web_dir):
    """

    Parameters
    ----------
    port : localhost port to start server on
    web_dir: directory containing server files

    Returns
    -------

    None

    """
    curr_dir = _os.path.abspath(_os.curdir)
    _os.chdir(web_dir)
    import subprocess
    import sys
    import webbrowser

    if port is None:
        port = _np.random.randint(8000, 9000)
    if sys.version_info[0] <= 2:
        subprocess.Popen(["python", "-m", "SimpleHTTPServer", str(port)])
    else:
        subprocess.Popen(["python", "-m", "http.server", str(port)])
    webbrowser.open_new_tab("http://localhost:{}".format(str(port)))
    _os.chdir(curr_dir)
    return True



from .builder import NeuralNetworkBuilder
from coremltools.models.utils import _get_model
import copy as _copy


def make_image_input(
    model,
    input_name,
    is_bgr=False,
    red_bias=0.0,
    blue_bias=0.0,
    green_bias=0.0,
    gray_bias=0.0,
    scale=1.0,
    image_format="NHWC",
):
    """
    Convert input of type multiarray to type image

    Parameters
    ----------
    TODO

    Returns
    -------
    model: MLModel
    A coreML MLModel object

    Examples
    --------
    TODO
    """

    spec = model.get_spec()

    if spec.WhichOneof("Type") not in [
        "neuralNetwork",
        "neuralNetworkClassifier",
        "neuralNetworkRegressor",
    ]:
        raise ValueError(
            "Provided model must be of type neuralNetwork, neuralNetworkClassifier or neuralNetworkRegressor"
        )

    if not isinstance(input_name, list):
        input_name = [input_name]

    spec_inputs = [i.name for i in spec.description.input]
    for name in input_name:
        if name not in spec_inputs:
            msg = "Provided input_name: {}, is not an existing input to the model"
            raise ValueError(msg.format(name))

    builder = NeuralNetworkBuilder(spec=spec)
    builder.set_pre_processing_parameters(
        image_input_names=input_name,
        is_bgr=is_bgr,
        red_bias=red_bias,
        green_bias=green_bias,
        blue_bias=blue_bias,
        gray_bias=gray_bias,
        image_scale=scale,
        image_format=image_format,
    )
    return _get_model(spec)


def make_nn_classifier(
    model,
    class_labels,
    predicted_feature_name=None,
    predicted_probabilities_output=None,
):
    """
    Convert a model of type "neuralNetwork" to type "neuralNetworkClassifier"

    Parameters
    ----------
    TODO

    Returns
    -------
    model: MLModel
    A coreML MLModel object

    Examples
    --------
    TODO
    """

    spec = model.get_spec()

    if spec.WhichOneof("Type") != "neuralNetwork":
        raise ValueError('Provided model must be of type "neuralNetwork"')

    # convert type to "neuralNetworkClassifier" and copy messages from "neuralNetwork"
    nn_spec = _copy.deepcopy(spec.neuralNetwork)
    spec.ClearField("neuralNetwork")
    for layer in nn_spec.layers:
        spec.neuralNetworkClassifier.layers.add().CopyFrom(layer)
    for preprocessing in nn_spec.preprocessing:
        spec.neuralNetworkClassifier.preprocessing.add().CopyFrom(preprocessing)
    spec.neuralNetworkClassifier.arrayInputShapeMapping = nn_spec.arrayInputShapeMapping
    spec.neuralNetworkClassifier.imageInputShapeMapping = nn_spec.imageInputShapeMapping
    spec.neuralNetworkClassifier.updateParams.CopyFrom(nn_spec.updateParams)

    # set properties related to classifier
    builder = NeuralNetworkBuilder(spec=spec)
    message = "Class labels must be a list of integers / strings or a file path"
    classes_in = class_labels
    if isinstance(classes_in, str):
        import os

        if not os.path.isfile(classes_in):
            raise ValueError("Path to class labels (%s) does not exist." % classes_in)
        with open(classes_in, "r") as f:
            classes = f.read()
        classes = classes.splitlines()
    elif isinstance(classes_in, list):  # list[int or str]
        classes = classes_in
        assert all([isinstance(x, \
            (int, str)) for x in classes]), message
    else:
        raise ValueError(message)

    kwargs = {}
    if predicted_feature_name is not None:
        kwargs["predicted_feature_name"] = predicted_feature_name
    if predicted_probabilities_output is not None:
        kwargs["prediction_blob"] = predicted_probabilities_output
    builder.set_class_labels(classes, **kwargs)

    return _get_model(spec)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil import types
from tensorflow.core.framework.types_pb2 import DataType
from tensorflow.python.framework.dtypes import _TF_TO_NP

import logging as _logging
import numpy as _np


def parse_type(t):
    mapping = {
        DataType.DT_HALF: types.fp16,
        DataType.DT_FLOAT: types.float,
        DataType.DT_DOUBLE: types.double,
        DataType.DT_INT32: types.int32,
        DataType.DT_UINT8: types.uint8,
        DataType.DT_INT16: types.int16,
        DataType.DT_INT8: types.int8,
        DataType.DT_STRING: types.str,
        DataType.DT_INT64: types.int64,
        DataType.DT_BOOL: types.bool,
        DataType.DT_UINT16: types.uint16,
        DataType.DT_UINT32: types.uint32,
        DataType.DT_UINT64: types.uint64,
    }
    t = int(t)
    if t in mapping:
        return mapping[t]
    else:
        _logging.info("Type %d cannot be mapped", t)
        return None


def parse_shape(t):
    if t.unknown_rank:
        return None
    ret = [d.size for d in t.dim]
    return ret


def parse_tensor(t):
    typ = parse_type(t.dtype)
    shape = parse_shape(t.tensor_shape)

    retval = None
    if len(t.half_val) > 0:
        retval = _np.array(t.half_val, dtype=_TF_TO_NP[t.dtype])
    elif len(t.float_val) > 0:
        retval = _np.array(t.float_val, dtype=_TF_TO_NP[t.dtype])
    elif len(t.double_val) > 0:
        retval = _np.array(t.double_val, dtype=_TF_TO_NP[t.dtype])
    elif len(t.int_val) > 0:
        retval = _np.array(t.int_val, dtype=_TF_TO_NP[t.dtype])
    elif len(t.int64_val) > 0:
        retval = _np.array(t.int64_val, dtype=_TF_TO_NP[t.dtype])
    elif len(t.bool_val) > 0:
        retval = _np.array(t.bool_val, dtype=_TF_TO_NP[t.dtype])
    elif hasattr(t, "uint32_val") and len(t.uint32_val) > 0:
        retval = _np.array(t.uint32_val, dtype=_TF_TO_NP[t.dtype])
    elif hasattr(t, "uint64_val") and len(t.uint64_val) > 0:
        retval = _np.array(t.uint64_val, dtype=_TF_TO_NP[t.dtype])

    if not t.tensor_shape.unknown_rank and len(shape) == 0:
        retobj = typ()
        if retval is not None:
            retobj.val = retval[0]
    else:
        rettype = types.tensor(typ, tuple(shape))
        retobj = rettype()
        retobj.shape = shape
        if retval is not None:
            retobj.val = retval

    return retobj


def parse_string(s):
    if isinstance(s, bytes):
        return s.decode("utf-8")
    else:
        return s


def parse_list(t):
    if len(t.s) > 0:
        return list(parse_string(s) for s in t.s)
    elif len(t.i) > 0:
        return list(t.i)
    elif len(t.f) > 0:
        return list(t.f)
    elif len(t.b) > 0:
        return list(t.b)
    elif len(t.type) > 0:
        return list(parse_type(z) for z in t.type)
    elif len(t.shape) > 0:
        return list(parse_shape(z) for z in t.shape)
    elif len(t.tensor) > 0:
        return list(parse_tensor(z) for z in t.tensor)
    else:
        return []


def parse_func(f):
    return f.name


def parse_attr(attr):
    if attr.HasField("s"):
        return parse_string(attr.s)
    elif attr.HasField("i"):
        return attr.i
    elif attr.HasField("f"):
        return attr.f
    elif attr.HasField("b"):
        return attr.b
    elif attr.HasField("type"):
        return parse_type(attr.type)
    elif attr.HasField("shape"):
        return parse_shape(attr.shape)
    elif attr.HasField("tensor"):
        return parse_tensor(attr.tensor)
    elif attr.HasField("list"):
        return parse_list(attr.list)
    elif attr.HasField("func"):
        return parse_func(attr.func)
    elif attr.HasField("placeholder"):
        raise NotImplementedError("placeholder not yet implemented")
    raise ValueError("unintelligible TFNode attributes")



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil import types
import numpy as np
import logging


@register_pass(namespace="tensorflow")
def expand_tf_lstm(prog):
    """
    Expand tf_lstm_block_cell to fine-grained SSA ops following:

    xh = [x, h_prev]
    [i, ci, f, o] = xh * w + b
    f = f + forget_bias
    if not use_peephole:
      wci = wcf = wco = 0
    i = sigmoid(cs_prev .* wci + i)
    f = sigmoid(cs_prev .* wcf + f)
    ci = tanh(ci)
    cs = ci .* i + cs_prev .* f
    cs = clip(cs, cell_clip)
    o = sigmoid(cs * wco + o)
    co = tanh(cs)
    h = co .* o

    Inputs:

        prog: Program
    """
    for f_name, f in prog.functions.items():
        expand_tf_lstm_helper(f)


def expand_tf_lstm_helper(block):
    # shallow copy hides changes on f.operations during the loop
    for op in block.operations[:]:
        for b in op.blocks:
            expand_tf_lstm_helper(b)

        if op.op_type == "tf_lstm_block_cell":
            expand_tf_lstm_block_cell(op)
            logging.info("Expanding {} (op_type: {})".format(op.name, op.op_type))

        if op.op_type == "tf_lstm_block":
            # only cs, h are supported for now. Can be easily extended to other outputs at performance hit.
            i, cs, f, o, ci, co, h = op.outputs
            if all(
                [
                    len(ov.child_ops) <= 0 and len(ov.consuming_blocks) <= 0
                    for ov in [i, f, o, ci, co]
                ]
            ):
                expand_tf_lstm_block(op)
                logging.info("Expanding {} (op_type: {})".format(op.name, op.op_type))


def _lstm_cell_builder(op, x, h_prev, cs_prev, before_op=None):
    b = op.bias  # [4*hidden_dim]
    forget_bias = op.forget_bias.val  # python:float

    # xh = [x, h_prev]
    # xh shape: [b, input_dim+hidden_dim]
    xh = mb.concat(values=[x, h_prev], axis=-1, before_op=before_op)

    # w: [4*hidden_dim, input_dim + hidden_dim] (icfo layout)
    w = np.transpose(op.weight.val)
    # [i, ci, f, o] = xh * w + b. Shape is [b, 4*hidden_dim]
    icfo = mb.linear(x=xh, weight=w, bias=b, before_op=before_op)

    # i, ci, f, o shape: [b, hidden_dim]
    i, ci, f, o = mb.split(x=icfo, num_splits=4, axis=-1, before_op=before_op)
    if op.forget_bias.val != 0:
        f = mb.add(x=f, y=forget_bias, before_op=before_op)

    # i = sigmoid(cs_prev .* wci + i)
    # f = sigmoid(cs_prev .* wcf + f)
    if op.use_peephole.val:
        wci = op.weight_peep_i.val  # [hidden_dim]
        wcf = op.weight_peep_f.val  # [hidden_dim]

        x = mb.mul(x=cs_prev, y=wci, before_op=before_op)
        pre_i = mb.add(x=x, y=i, before_op=before_op)

        x = mb.mul(x=cs_prev, y=wcf, before_op=before_op)
        pre_f = mb.add(x=x, y=f, before_op=before_op)
    else:
        pre_i = i
        pre_f = f

    i = mb.sigmoid(x=pre_i, before_op=before_op)
    f = mb.sigmoid(x=pre_f, before_op=before_op)

    # ci = tanh(ci)
    ci = mb.tanh(x=ci, before_op=before_op)

    # cs = ci .* i + cs_prev .* f
    x = mb.mul(x=ci, y=i, before_op=before_op)
    y = mb.mul(x=cs_prev, y=f, before_op=before_op)
    cs = mb.add(x=x, y=y, before_op=before_op)

    # cs = clip(cs, cell_clip)
    if op.cell_clip is not None:
        clip_val = op.cell_clip.val
        cs = mb.clip(x=cs, alpha=-clip_val, beta=clip_val, before_op=before_op)

    # o = sigmoid(cs * wco + o)
    if op.use_peephole.val:
        wco = op.weight_peep_o.val
        x = mb.mul(x=cs, y=wco, before_op=before_op)
        pre_o = mb.add(x=x, y=o, before_op=before_op)
    else:
        pre_o = o
    o = mb.sigmoid(x=pre_o, before_op=before_op)

    # co = tanh(cs)
    co = mb.tanh(x=cs, before_op=before_op)

    # h = co .* o
    h = mb.mul(x=co, y=o, before_op=before_op)

    return [i, cs, f, o, ci, co, h]


def expand_tf_lstm_block_cell(op):
    if op.op_type != "tf_lstm_block_cell":
        raise ValueError()

    with op.enclosing_block as block:
        x = op.x  # [b, input_dim]
        h_prev = op.h_prev  # [b, hidden_dim]
        cs_prev = op.c_prev  # [b, hidden_dim]

        i, cs, f, o, ci, co, h = _lstm_cell_builder(
            op, x, h_prev, cs_prev, before_op=op
        )

        # Replace all outputs
        new_outputs = [i, cs, f, o, ci, co, h]
        for old_v, new_v in zip(op.outputs, new_outputs):
            block.replace_uses_of_var_after_op(
                anchor_op=op, old_var=old_v, new_var=new_v
            )
        block.remove_ops([op])


def expand_tf_lstm_block(op):
    if op.op_type != "tf_lstm_block":
        raise ValueError()

    with op.enclosing_block as block:
        x = op.x  # [s, b, input_dim]
        h_prev = op.h_prev  # [b, hidden_dim]
        cs_prev = op.c_prev  # [b, hidden_dim]

        # Allocate two lists: cs & h
        x_shape = mb.shape(x=x, before_op=op)
        length = mb.slice_by_index(x=x_shape, begin=[0], end=[1], before_op=op)
        h_shape = mb.shape(x=h_prev, before_op=op)
        list_shape = mb.concat(values=[length, h_shape], axis=0, before_op=op)
        cs_list = mb.fill(shape=list_shape, before_op=op)
        h_list = mb.fill(shape=list_shape, before_op=op)

        # append initial state at index 0
        cs_prev = mb.expand_dims(x=cs_prev, axes=[0], before_op=op)
        cs_list = mb.concat(values=[cs_prev, cs_list], axis=0, before_op=op)
        h_prev = mb.expand_dims(x=h_prev, axes=[0], before_op=op)
        h_list = mb.concat(values=[h_prev, h_list], axis=0, before_op=op)

        def cond(i, cs_list, h_list):
            return mb.less(x=i, y=length)

        def body(i, cs_list, h_list):
            xi = mb.gather(x=x, indices=i, axis=0)
            h_prev = mb.gather(x=h_list, indices=i, axis=0)
            cs_prev = mb.gather(x=cs_list, indices=i, axis=0)

            ig, cs, fg, og, ci, co, h = _lstm_cell_builder(op, xi, h_prev, cs_prev)

            counter = mb.add(x=i, y=1)

            return (
                counter,
                mb.scatter(data=cs_list, indices=counter, updates=cs),
                mb.scatter(data=h_list, indices=counter, updates=h),
            )

        _, cs_list, h_list = mb.while_loop(
            _cond=cond, _body=body, loop_vars=([0], cs_list, h_list), before_op=op
        )

        # strip initial state or element at index 0
        begin, end = [1, 0, 0], [0, 0, 0]
        begin_mask = [False, True, True]
        end_mask = [True, True, True]
        cs = mb.slice_by_index(
            x=cs_list,
            begin=begin,
            end=end,
            begin_mask=begin_mask,
            end_mask=end_mask,
            before_op=op,
        )
        h = mb.slice_by_index(
            x=h_list,
            begin=begin,
            end=end,
            begin_mask=begin_mask,
            end_mask=end_mask,
            before_op=op,
        )

        # Replace all outputs
        new_outputs = [cs, h]
        for old_v, new_v in zip(
            [ov for index, ov in enumerate(op.outputs) if index in [1, 6]], new_outputs
        ):
            block.replace_uses_of_var_after_op(
                anchor_op=op, old_var=old_v, new_var=new_v
            )
        block.remove_ops([op])



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil.types.symbolic import any_symbolic


@register_pass(namespace="common")
def add_conv_transpose_output_shape(prog):
    """
    ``conv_transpose`` input ``output_shape`` is an optional input.
    Since we can infer the output shape from type_inference, we add
    ``output_shape`` input whenever it is known to be constant at
    compile time. Ex:

    Given:
      %1: (1, 5, 39, fp32) = conv_transpose(...) # no output_shape input.

    Result:
      %2: (3, i32) = const(val=[1,5,39])
      %3: (1, 5, 39, fp32) = conv_transpose(..., output_shape=%2)
    """
    for f_name, f in prog.functions.items():
        handle_block(f)

def match_pattern(op):
  return op.op_type == "conv_transpose" \
      and op.output_shape is None \
      and not any_symbolic(op.outputs[0].shape)

def handle_block(block):
    for op in list(block.operations):
        for b in op.blocks:
            handle_block(b)

        if not match_pattern(op):
            continue

        # matched pattern
        with block:
            x = mb.conv_transpose(
                **op.inputs, output_shape=op.outputs[0].shape, \
                name=op.name+'_has_output_shape', before_op=op
            )
            op.enclosing_block.replace_uses_of_var_after_op(
                anchor_op=op, old_var=op.outputs[0], new_var=x
            )
            block.remove_ops([op])



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from . import datatypes
from .. import SPECIFICATION_VERSION
from ..proto import Model_pb2 as _Model_pb2
from ..proto import FeatureTypes_pb2 as _FeatureTypes_pb2
from ._interface_management import set_transform_interface_params


def create_array_feature_extractor(
    input_features, output_name, extract_indices, output_type=None
):
    """
    Creates a feature extractor from an input array feature, return

    input_features is a list of one (name, array) tuple.

    extract_indices is either an integer or a list.  If it's an integer,
    the output type is by default a double (but may also be an integer).
    If a list, the output type is an array.
    """

    # Make sure that our starting stuff is in the proper form.
    assert len(input_features) == 1
    assert isinstance(input_features[0][1], datatypes.Array)

    # Create the model.
    spec = _Model_pb2.Model()
    spec.specificationVersion = SPECIFICATION_VERSION

    if isinstance(extract_indices, int):
        extract_indices = [extract_indices]
        if output_type is None:
            output_type = datatypes.Double()

    elif isinstance(extract_indices, (list, tuple)):
        if not all(isinstance(x, int) for x in extract_indices):
            raise TypeError("extract_indices must be an integer or a list of integers.")

        if output_type is None:
            output_type = datatypes.Array(len(extract_indices))

    else:
        raise TypeError("extract_indices must be an integer or a list of integers.")

    output_features = [(output_name, output_type)]

    for idx in extract_indices:
        assert idx < input_features[0][1].num_elements
        spec.arrayFeatureExtractor.extractIndex.append(idx)

    set_transform_interface_params(spec, input_features, output_features)

    return spec



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.frontend.tensorflow.test import (
    testing_utils as tf_testing_utils,
)
from coremltools.converters.mil.frontend.tensorflow2.test.testing_utils import (
    make_tf2_graph as make_tf_graph
)
from coremltools.converters.mil.testing_reqs import *

tf = pytest.importorskip("tensorflow", minversion="2.1.0")

backends = testing_reqs.backends

# -----------------------------------------------------------------------------
# Overwrite utilities to enable different conversion / compare method
tf_testing_utils.frontend = "TensorFlow2"
tf_testing_utils.make_tf_graph = make_tf_graph

# -----------------------------------------------------------------------------
# Import TF 2.x-compatible TF 1.x test cases
from coremltools.converters.mil.frontend.tensorflow.test.test_composite_ops import (
    TestCompositeOp,
)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
import copy
import logging

from coremltools.converters.mil.mil import types
from .basic_graph_ops import check_connections, const_determined_nodes
from .dot_visitor import DotVisitor
from .naming_utils import escape_fn_name


class ParsedNode(object):
    """
    Node class for the tfssa graph.

    name: The name of the node (str)
    op: The operation represented by the node (str)
    datatype: The type of the node. (type)
    value: The value of the node if available
    inputs: The list of nodes which are inputs to this node (list[str])
    control_inputs: The list of nodes which have to be executed before this node (list[str])
    attr: The attributes of the node
    outputs: The list of nodes which consume the result of this node (list[str])
    control_outputs: The list of nodes which have to be executed after this node (list[str])
    """

    __slots__ = [
        "name",
        "op",
        "datatype",
        "value",
        "inputs",
        "control_inputs",
        "outputs",
        "control_outputs",
        "attr",
    ]

    def __init__(self):
        self.name = None
        self.op = None
        self.datatype = None
        self.value = None
        self.inputs = []
        self.outputs = []
        self.control_inputs = []
        self.control_outputs = []
        self.attr = {}

    def __copy__(self):
        return self._copy_impl(ParsedNode())

    def _copy_impl(self, dest):
        dest.name = self.name
        dest.op = self.op
        dest.datatype = self.datatype
        dest.value = copy.deepcopy(self.value)
        dest.inputs = self.inputs[:]
        dest.control_inputs = self.control_inputs[:]
        dest.outputs = self.outputs[:]
        dest.control_outputs = self.control_outputs[:]
        dest.attr = {k: copy.deepcopy(v) for k, v in self.attr.items()}
        return dest

    def copy(self):
        return self.__copy__()


class SSAFunction(object):
    __slots__ = ["graph", "inputs", "input_types", "outputs", "output_types", "ret"]

    def __init__(self, gdict=None, inputs=None, outputs=None, ret=None):
        if gdict is None:
            gdict = {}
        self.graph = gdict
        self.inputs = [] if inputs is None else inputs
        self.outputs = [] if outputs is None else outputs
        self.input_types = []
        self.output_types = []

        # ret is a mapping from the output arg names from `signature` to the
        # outputs from `node_def` that should be returned by the function.
        # Only used in TF2 for getting indices when generating get_tuple ops
        # for control flow ops. Because the sub-graph's outputs and control
        # flow node's outputs mapping is defined in `ret` dict. See usages in
        # tf_graph_pass: rewrite_control_flow_functions for details.
        # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/function.proto
        self.ret = [] if ret is None else ret

        check_connections(gdict)

        # respect TF inputs/outputs if given, otherwise, infer from the graph
        # in currently implementation: TF1 will always infer from graph. TF2,
        # on the other hand, respect the inputs/outputs provided.
        if len(self.inputs) == 0 or len(self.outputs) == 0:
            self.find_inputs_and_outputs()
        else:
            self.inputs, self.outputs = inputs, outputs
            self.filter_inputs_and_outputs()

    def find_inputs_and_outputs(self):
        # solve for input and output vars
        sorted_keys = sorted(self.graph.keys())

        # we use function entry and exit points if available
        # otherwise we find graph entry and exit points
        # TODO: op name should be fixed here.
        #       <rdar://problem/57081966> Remove wrappers that are used for old tfssa
        enters = [
            n.name for n in self.graph.values() if ("entry" in n.op or "Entry" in n.op)
        ]
        exits = [n.name for n in self.graph.values() if n.op in ("Return", "return")]
        if len(enters) > 0 or len(exits) > 0:
            assert len(enters) > 0
            assert len(exits) > 0
            self.inputs = enters
            self.input_types = [self.graph[v].datatype for v in self.inputs]
            self.outputs = exits
            self.output_types = [self.graph[v].datatype for v in self.outputs]
        else:
            for k in sorted_keys:
                v = self.graph[k]
                if len(v.inputs) == 0 and v.op not in ["Const", "get_global", "NoOp"]:
                    self.inputs.append(k)
                    self.input_types.append(v.datatype)
                elif len(v.inputs) != 0 and v.op == "Placeholder":
                    assert len(v.inputs) == 1, "This is not a PlaceholderWithDefault!"
                    self.inputs.append(k)
                    self.input_types.append(v.datatype)
                if (
                    len(v.outputs) == 0
                    and len(v.control_outputs) == 0
                    and v.op != "set_global"
                ):
                    self.outputs.append(k)
                    self.output_types.append(v.datatype)

    def filter_inputs_and_outputs(self):
        """
        Eliminate invalid input/output nodes in the given list. Should only be
        invoked if the self.inputs and self.outputs are both provided and we
        want to respect those when adding SSAFunctions. Only needed for TF2 for
        now because of the needs to parse multiple functions in graph. TF1 only
        has one "main" function.
        """
        filtered_inputs = []
        filtered_outputs = []
        for k in self.inputs:
            if k not in self.graph.keys():
                continue
            v = self.graph[k]
            if len(v.inputs) == 0 and v.op not in {"Const", "get_global", "NoOp"}:
                filtered_inputs.append(k)
                self.input_types.append(v.datatype)
            elif len(v.inputs) != 0 and v.op == "Placeholder":
                assert len(v.inputs) == 1, "This is not a PlaceholderWithDefault!"
                filtered_inputs.append(k)
                self.input_types.append(v.datatype)
        for k in self.outputs:
            if k not in self.graph.keys():
                continue
            v = self.graph[k]
            filtered_outputs.append(k)
            self.output_types.append(v.datatype)
        self.inputs, self.outputs = filtered_inputs, filtered_outputs

    def __copy__(self):
        ret = SSAFunction()
        ret.inputs = self.inputs[:]
        ret.input_types = self.input_types[:]
        ret.outputs = self.outputs[:]
        ret.output_types = self.output_types[:]
        ret.graph = {k: copy.deepcopy(v) for k, v in self.graph.items()}

        return ret

    def copy(self):
        return self.__copy__()


class NetworkEnsemble(object):
    __slots__ = ["functions", "variables", "global_resource"]

    def __init__(self, instance=None):
        self.functions = {}
        self.variables = {}
        self.global_resource = {}

        if isinstance(instance, NetworkEnsemble):
            self.functions = instance.functions
            self.variables = instance.variables
            self.global_resource = instance.global_resource
        elif instance is not None:
            raise ValueError(
                "Instance type {} not compatible with NetworkEnsemble".format(
                    type(instance)
                )
            )

    def rename_function(self, src_func, tgt_func):
        """
        Renames the function with function name (src_func) to (tgt_func)
        """
        if src_func not in self.functions:
            logging.warning("Couldn't find function name (%s).", src_func)
            return
        if tgt_func in self.functions:
            logging.warning("(%s) already exists in some function name.", tgt_func)
            return

        self.functions[tgt_func] = self.functions.pop(src_func)
        logging.debug(
            "Successfully changed function name from (%s) to (%s)", src_func, tgt_func
        )

    def rename_node(self, src_node, tgt_node):
        """
        Rename the node with node name (src_node) to (tgt_node).
        Note that the name (tgt_node) cannot appear in the whole network,
        not only the function it lies in.
        """
        in_ssa = False
        success = None
        for func, tfssa in self.functions.items():
            if src_node in tfssa.graph:
                in_ssa = True
                if tgt_node in tfssa.graph:
                    logging.warning(
                        "(%s) already exists in function (%s).", tgt_node, func
                    )
                    break
                success = func
                tfssa.graph[tgt_node] = tfssa.graph.pop(src_node)
                # Replace other nodes' output dependency
                for inp in tfssa.graph[tgt_node].inputs:
                    for idx, out in enumerate(tfssa.graph[inp].outputs):
                        if out == src_node:
                            tfssa.graph[inp].outputs[idx] = tgt_node
                            break
                # Replace other nodes' control output dependency
                for c_inp in tfssa.graph[tgt_node].control_inputs:
                    for idx, c_out in enumerate(tfssa.graph[c_inp].control_outputs):
                        if c_out == src_node:
                            tfssa.graph[c_inp].control_outputs[idx] = tgt_node
                            break
                # Replace other nodes' input dependency
                for out in tfssa.graph[tgt_node].outputs:
                    for idx, inp in enumerate(tfssa.graph[out].inputs):
                        if inp == src_node:
                            tfssa.graph[out].inputs[idx] = tgt_node
                            break
                # Replace other nodes' control input dependency
                for c_out in tfssa.graph[tgt_node].control_outputs:
                    for idx, c_inp in enumerate(tfssa.graph[c_out].control_inputs):
                        if c_inp == src_node:
                            tfssa.graph[c_out].control_inputs[idx] = tgt_node
                            break
                break

        if not in_ssa:
            logging.warning("Couldn't find (%s) in any functions", src_node)
        if success is not None:
            logging.debug(
                "Changed (%s) to (%s) in function (%s)", src_node, tgt_node, success
            )

    def extract_subgraph(self, outputs, target_inputs=None, name=""):
        """Add a new SSAFunction to the current NetworkEnsemble to produce the given outputs.

        Args:
            outputs: The outputs the new function must produce.
            target_inputs:
            name: The name of the new function to create. If unspecified, a name will be generated
                  by joining output names.
        Returns:
            The name of the new function.
        """
        if not isinstance(outputs, list):
            raise TypeError("Expected a list of output names for subgraph extraction")

        if name == "":
            outputs.sort()
            name = escape_fn_name("_".join(outputs))

        if target_inputs is None:
            target_inputs = []

        def DFS_inputs(graph, node, vis):
            vis.add(node)
            if node in target_inputs:
                return [node]
            if (
                len(graph[node].inputs) == 0
                and len(graph[node].control_inputs) == 0
                and graph[node].op != "Const"
            ):
                return [node]
            inputs = []
            for i in graph[node].inputs + graph[node].control_inputs:
                if i in vis:
                    continue
                inputs += DFS_inputs(graph, i, vis)
            return inputs

        def DFS_set_globals(graph, node, vis):
            vis.add(node)
            set_globals = []
            if graph[node].op == "set_global":
                set_globals.append(node)
            for i in graph[node].outputs + graph[node].control_outputs:
                if i in vis:
                    continue
                set_globals += DFS_set_globals(graph, i, vis)
            return set_globals

        for k in list(self.functions.keys()):
            v = self.functions[k]
            extract = []
            for output in outputs:
                if output in v.graph:
                    extract.append(output)

            if len(extract) == 0:
                continue
            incl_nodes = set()
            gdict = copy.deepcopy(v.graph)
            inputs = []
            set_globals = []
            for output in extract:
                inputs += DFS_inputs(gdict, output, incl_nodes)
            vis_nodes = set()
            for inp in inputs:
                set_globals += DFS_set_globals(gdict, inp, vis_nodes)
            for node in set_globals:
                inputs += DFS_inputs(gdict, node, incl_nodes)

            for new_k, new_v in v.graph.items():
                if new_k not in incl_nodes:
                    del gdict[new_k]
                    continue
                if new_k in target_inputs:
                    gdict[new_k].op = "Placeholder"
                gdict[new_k].inputs = [inp for inp in new_v.inputs if inp in incl_nodes]
                gdict[new_k].outputs = [
                    out for out in new_v.outputs if out in incl_nodes
                ]
                gdict[new_k].control_inputs = [
                    inp for inp in new_v.control_inputs if inp in incl_nodes
                ]
                gdict[new_k].control_outputs = [
                    out for out in new_v.control_outputs if out in incl_nodes
                ]

            for output in extract:
                old_name = "preIdentity_" + output
                output_node = copy.deepcopy(gdict[output])
                output_node.op = "Identity"
                output_node.inputs = [old_name]
                output_node.control_inputs = []
                output_node.outputs = []
                output_node.control_outputs = []

                for inp in gdict[output].inputs:
                    for idx, out in enumerate(gdict[inp].outputs):
                        if out == output:
                            gdict[inp].outputs[idx] = old_name
                for inp in gdict[output].control_inputs:
                    for idx, out in enumerate(gdict[inp].control_outputs):
                        if out == output:
                            gdict[inp].control_outputs[idx] = old_name
                for out in gdict[output].outputs:
                    for idx, inp in enumerate(gdict[out].inputs):
                        if inp == output:
                            gdict[out].inputs[idx] = old_name
                for out in gdict[output].control_outputs:
                    for idx, inp in enumerate(gdict[out].control_inputs):
                        if inp == output:
                            gdict[out].control_inputs[idx] = old_name
                gdict[output].outputs.append(output)
                gdict[output].name = old_name
                gdict[old_name] = gdict[output]
                gdict[output] = output_node

            self.functions[name] = SSAFunction(gdict)
        return name

    def delete_subgraph(self, name):
        """
        Delete the SSAfunction with function_name.
        """
        if name not in self.functions:
            logging.warning("(%s) not in NetworkEnsemble", name)
            return
        del self.functions[name]

    def __repr__(self):
        return str(self)

    def __str__(self):
        ret = ""
        for func, v in self.functions.items():
            if func.startswith("body_function_") or func.startswith("f_body_function_"):
                continue
            elif func.startswith("cond_function_") or func.startswith(
                "f_cond_function_"
            ):
                continue

            ret += "Input Function Name: %s\n" % (func)
            ret += "  Inputs:\n"
            for inp in v.inputs:
                ret += "    %s\n" % (inp)
            ret += "  Outputs:\n"
            for out in v.outputs:
                if out.startswith("fake_exit_"):
                    continue
                ret += "    %s\n" % (out)
        return ret

    def get_dot_string(
        self, name_and_op_style=False, annotation=False, highlight_debug_nodes=None
    ):
        """
        Return the dot string that can be used to show the whole graph
        with dot. By default, the graph contains op and type. If
        name_and_op_style is set, the graph will contain the name of the node
        and the op instead.

        * Input nodes : yellow
        * constant nodes : azure
        * output nodes : goldenrod2
        * nodes with variable shaped tensors : cyan
        * node names or op types that user wants to highlight: green

        Parameters
        ----------
        name_and_op_style: bool
            If set, graph contains only the name and the op.

        annotation: bool
        Examples
        --------
        >>> import graphviz
        >>> graphviz.Source(network.get_dot_string()).view()

        """
        if highlight_debug_nodes is None:
            highlight_debug_nodes = []
        function_names = sorted(self.functions.keys())

        dotstring = "digraph g {\n" + "\tcompound=true;\n"
        # find all tensor nodes with unknown sizes
        ctr = 0
        for k in function_names:
            const_nodes = const_determined_nodes(self.functions[k].graph)
            unknown_sized_tensor_ops = []
            for v, n in self.functions[k].graph.items():
                if n.datatype is None or (
                    n.datatype is not None
                    and types.is_tensor(n.datatype)
                    and (
                        len(n.datatype.get_shape()) == 0 or -1 in n.datatype.get_shape()
                    )
                ):
                    unknown_sized_tensor_ops.append(v)
                if n.op in highlight_debug_nodes:
                    highlight_debug_nodes.append(v)

            v = self.functions[k]
            vis = DotVisitor(annotation)
            vis.highlight_nodes(v.inputs, "yellow").highlight_nodes(
                const_nodes, "azure2"
            ).highlight_nodes(v.outputs, "goldenrod2").highlight_nodes(
                unknown_sized_tensor_ops, "cyan2"
            )
            if len(highlight_debug_nodes) > 0:
                vis.highlight_nodes(highlight_debug_nodes, "green")
            if name_and_op_style:
                vis.labeller(lambda n: n.name + " (" + n.op + ")")

            res = vis.visit_all(v.graph, nodename_prefix=str(ctr)).get_result(
                "subgraph", "cluster_" + k.replace("/", "_")
            )
            dotstring += "\n".join("\t" + r for r in res.split("\n")) + "\n"
            ctr += 1
        dotstring += "}"
        return dotstring

    def add_function_with_prefix(self, fprefix, tfssa):
        assert isinstance(tfssa, SSAFunction)
        s = 0
        while fprefix + str(s) in self.functions:
            s += 1
        self.functions[fprefix + str(s)] = tfssa

    def add_function(self, f, tfssa):
        self.functions[f] = tfssa

    def __copy__(self):
        ret = self.__class__()
        ret.functions = self.functions
        ret.variables = self.variables
        ret.global_resource = self.global_resource
        return ret

    def __deepcopy__(self, memo):
        ret = self.__class__()
        ret.functions = {k: copy.copy(v) for k, v in self.functions.items()}
        ret.variables = {k: copy.copy(v) for k, v in self.variables.items()}
        ret.global_resource = {k: copy.copy(v) for k, v in self.global_resource.items()}
        return ret

    def copy(self):
        return self.__copy__()

    def _find_free_name(self, prefix):
        idx = 0
        while True:
            name = prefix + str(idx)
            found = False
            for v in self.functions.values():
                if name in v.graph:
                    found = True
                    break
            if found:
                idx += 1
            else:
                return name

    def get_image_format(self):
        """
        Iterates over graph and returns input format (`NCHW` or `NHWC`)
        if input is of type Image, otherwise `None`
        """
        for fn_key in list(self.functions.keys()):
            graph = self.functions[fn_key].graph

            for name in graph:
                node = graph[name]
                if (
                    node.attr.get("data_format", None) == "NHWC"
                    or node.attr.get("data_format") == "NHWC_format_inserted"
                ):
                    return "NHWC"
                elif node.attr.get("data_format", None) == "NCHW":
                    return "NCHW"
        return None



from __future__ import absolute_import as _

import gc
import coremltools
import collections

from coremltools.converters.mil.input_types import InputType, ClassifierConfig
from coremltools.converters.mil.converter import mil_convert
from coremltools.converters.mil.mil import Program
from coremltools._deps import _HAS_TORCH, _HAS_TF_1, _HAS_TF_2
from coremltools.converters._profile_utils import _profile
from coremltools import __version__ as ct_version
from coremltools.models import _METADATA_VERSION, _METADATA_SOURCE
from coremltools.converters.mil._deployment_compatibility import (
    AvailableTarget,
    check_deployment_compatibility,
)

if _HAS_TF_1:
    import tensorflow as tf
    from coremltools.converters.mil.frontend.tensorflow.load import TF1Loader
if _HAS_TF_2:
    import tensorflow as tf
    from coremltools.converters.mil.frontend.tensorflow2.load import TF2Loader

if _HAS_TORCH:
    import torch
    from coremltools.converters.mil.frontend.torch.load import (
        _torchscript_from_model as pytorch_load,
    )


@_profile
def convert(
    model,
    source="auto",
    inputs=None,
    outputs=None,
    classifier_config=None,
    minimum_deployment_target=None,
    convert_to='nn_proto',
    **kwargs
):
    """
    Convert TensorFlow or Pytorch models to the Core ML model format. Whether a
    parameter is required may differ between frameworks (see below). Note that
    this function is aliased as `ct.convert` in the tutorials.

    Parameters
    ----------
    model:
        TensorFlow 1, TensorFlow 2 or Pytorch model in one of the following
        formats:

        For TensorFlow versions 1.x:
            - Frozen `tf.Graph <https://www.tensorflow.org/api_docs/python/tf/Graph>`_
            - Frozen graph (``.pb``) file path
            - `tf.keras.Model <https://www.tensorflow.org/api_docs/python/tf/keras>`_
            -  `HDF5 <https://keras.io/api/models/model_saving_apis/>`_ file path (``.h5``)
            - `SavedModel <https://www.tensorflow.org/guide/saved_model>`_ directory path
        For TensorFlow versions 2.x:
            - `tf.keras.Model <https://www.tensorflow.org/api_docs/python/tf/keras>`_
            - `HDF5 file path <https://keras.io/api/models/model_saving_apis/>`_ (``.h5``)
            - `SavedModel <https://www.tensorflow.org/guide/saved_model>`_ directory path
            - A `concrete function <https://www.tensorflow.org/guide/concrete_function>`_
        For Pytorch:
            - A `TorchScript <https://pytorch.org/docs/stable/jit.html>`_ object
            - Path to a ``.pt`` file

    source: str (optional)
        One of [``auto``, ``tensorflow``, ``pytorch``, ``mil``]. `auto` determines the
        framework automatically for most cases. Raise ValueError if it fails
        to determine the source framework.

    inputs: list of `TensorType` or `ImageType`
        TensorFlow 1 and 2:
            - ``inputs`` are optional. If not provided, the inputs are
              Placeholder nodes in the model (if model is frozen graph) or
              function inputs (if model is tf function)
            - ``inputs`` must corresponds to all or some of the Placeholder
              nodes in the TF model
            - ``TensorType`` and ``ImageType`` in ``inputs`` must have ``name``
              specified. ``shape`` is optional.
            - If ``inputs`` is provided, it must be a flat list.

        PyTorch:
            - ``inputs`` are required.
            - ``inputs`` may be nested list or tuple.
            - ``TensorType`` and ``ImageType`` in ``inputs`` must have ``name``
              and ``shape`` specified.

    outputs: list[str] (optional)

        TensorFlow 1 and 2:
            - ``outputs`` are optional.

            - If specified, ``outputs`` is a list of string representing node
              names.

            - If ``outputs`` are not specified, converter infers outputs as all
              terminal identity nodes.

        PyTorch:
            - ``outputs`` must not be specified.

    classifier_config: ClassifierConfig class (optional)
        The configuration if the mlmodel is intended to be a classifier.

    minimum_deployment_target: coremltools.target enumeration (optional)
        - one of the members of enum ``coremltools.target``.
        - When not-specified or None, converter aims for as minimum of a
          deployment target as possible

    convert_to: str (optional)
        - Must be one of [``'nn_proto'``, ``'mil'``].
        - ``'nn_proto'``: Returns MLModel containing a NeuralNetwork
          proto
        - ``'mil'``: Returns MIL program object. MIL program is primarily used
          for debugging purpose and currently cannot be compiled to
          executable.

    Returns
    -------
    model: `coremltools.models.MLModel` or `coremltools.converters.mil.Program`
        A Core ML MLModel object or MIL Program object (see ``convert_to``)

    Examples
    --------
    TensorFlow 1, 2 (``model`` is a frozen graph):

        >>> with tf.Graph().as_default() as graph:
        >>>     x = tf.placeholder(tf.float32, shape=(1, 2, 3), name="input")
        >>>     y = tf.nn.relu(x, name="output")

        # Automatically infer inputs and outputs

        >>> mlmodel = ct.convert(graph)
        >>> test_input = np.random.rand(1, 2, 3) - 0.5
        >>> results = mlmodel.predict({"input": test_input})
        >>> print(results['output'])

    TensorFlow 2 (``model`` is tf.Keras model path):

        >>> x = tf.keras.Input(shape=(32,), name='input')
        >>> y = tf.keras.layers.Dense(16, activation='softmax')(x)
        >>> keras_model = tf.keras.Model(x, y)

        >>> keras_model.save(h5_path)
        >>> mlmodel = ct.convert(h5_path)

        >>> test_input = np.random.rand(2, 32)
        >>> results = mlmodel.predict({'input': test_input})
        >>> print(results['Identity'])

    Pytorch:

        >>> model = torchvision.models.mobilenet_v2()
        >>> model.eval()
        >>> example_input = torch.rand(1, 3, 256, 256)
        >>> traced_model = torch.jit.trace(model, example_input)

        >>> input = ct.TensorType(name='input_name', shape=(1, 3, 256, 256))
        >>> mlmodel = ct.convert(traced_model, inputs=[input])
        >>> results = mlmodel.predict({"input": example_input.numpy()})
        >>> print(results['1651']) # 1651 is the node name given by PyTorch's JIT

    See `neural-network-conversion <https://coremltools.readme.io/docs/neural-network-conversion>`_ for
    more advanced options.
    """
    _check_deployment_target(minimum_deployment_target)
    exact_source = _determine_source(model, source, outputs)
    _validate_inputs(model, exact_source, inputs, outputs, classifier_config,
        **kwargs)

    mlmodel = mil_convert(
        model,
        convert_from=exact_source,
        convert_to=convert_to,
        inputs=inputs,
        outputs=outputs,
        classifier_config=classifier_config,
        **kwargs
    )

    if convert_to == 'mil':
        return mlmodel # Returns the MIL program

    if minimum_deployment_target is not None:
        check_deployment_compatibility(
            spec=mlmodel.get_spec(),
            representation=convert_to,
            deployment_target=minimum_deployment_target,
        )

    gc.collect()

    mlmodel = _record_src_version(mlmodel, exact_source)
    mlmodel.user_defined_metadata[_METADATA_VERSION] = ct_version

    return mlmodel


def _check_deployment_target(minimum_deployment_target):
    if minimum_deployment_target is not None and \
        not isinstance(minimum_deployment_target, AvailableTarget):
        msg = (
            "Unrecognized value of argument 'minimum_deployment_target': {}. "
            "It needs to be a member of 'coremltools.target' enumeration. "
            "For example, coremltools.target.iOS13"
        )
        raise TypeError(msg.format(minimum_deployment_target))

def _validate_inputs(model, exact_source, inputs, outputs, classifier_config,
    **kwargs):
    """
    Validate and process model, inputs, outputs, classifier_config based on
    `exact_source` (which cannot be `auto`)
    """
    def raise_if_duplicated(input_list):
        # Detect duplicated inputs
        input_names = [t.name for t in input_list if t.name is not None]
        dups = [
            item
            for item, count in collections.Counter(input_names).items()
            if count > 1
        ]
        if len(dups) > 0:
            raise ValueError("Duplicated inputs: {}".format(dups))

    if inputs is not None:
        if not isinstance(inputs, list):
            msg = '"inputs" must be of type list'
            raise ValueError(msg)

    if classifier_config is not None:
        if not isinstance(classifier_config, ClassifierConfig):
            msg = '"classifier_config" must be of type ClassifierConfig'
            raise ValueError(msg)

    if exact_source in {"tensorflow", "tensorflow2"}:
        if exact_source == "tensorflow" and not _HAS_TF_1:
            msg = 'Converter was called with source="tensorflow", ' +\
                    'but missing tensorflow package'
            raise ValueError(msg)

        if inputs is not None:
            raise_if_duplicated(inputs)

        if inputs is not None and not all(
            [isinstance(_input, InputType) for _input in inputs]
        ):
            raise ValueError("Input should be a list of TensorType or ImageType")

    elif exact_source == "pytorch":
        if "example_inputs" in kwargs:
            msg = 'Unexpected argument "example_inputs" found'
            raise ValueError(msg)

        if inputs is None:
            msg = 'Expected argument for pytorch "inputs" not provided'
            raise ValueError(msg)

        def _flatten_list(_inputs):
            ret = []
            for _input in _inputs:
                if isinstance(_input, (list, tuple)):
                    ret.extend(_flatten_list(_input))
                elif isinstance(_input, InputType):
                    ret.append(_input)
                else:
                    raise ValueError(
                        "Unknown type {} for flattening into InputType.".format(
                            type(_input)
                        )
                    )
            return ret

        flat_inputs = _flatten_list(inputs)
        raise_if_duplicated(flat_inputs)
        if inputs is not None and not all(
            [isinstance(_input, InputType) for _input in flat_inputs]
        ):
            raise ValueError(
                "Input should be a list/tuple (or nested lists/tuples) of TensorType or ImageType"
            )
        if outputs is not None:
            raise ValueError("outputs must not be specified for PyTorch")

    elif exact_source == "mil":
        if not isinstance(model, Program):
            msg = "Converter was asked to convert MIL input, but input is not a MIL program!"
            raise ValueError(msg)


def _determine_source(model, source, outputs):
    """
    Infer source (which can be auto) to the precise framework.
    """
    source = source.lower()
    if source not in {"auto", "tensorflow", "pytorch", "mil"}:
        msg = (
            'Unrecognized value of argument "source": {}. '
            'It must be one of ["auto", "tensorflow", "pytorch"].'
        )
        raise ValueError(msg.format(source))


    # Determine tensorflow version
    if source == "tensorflow" and _HAS_TF_2:
        return "tensorflow2"

    if source != 'auto':
        return source

    # Determine `auto` source
    if source == "auto" and _HAS_TF_1:
        try:
            loader = TF1Loader(model, outputs=outputs)
            loader._graph_def_from_model(outputs=outputs)
            return "tensorflow"
        except:
            pass

    if source == "auto" and _HAS_TF_2:
        try:
            loader = TF2Loader(model, outputs=outputs)
            loader._graph_def_from_model(outputs=outputs)
            return "tensorflow2"
        except:
            pass

    if source == "auto" and _HAS_TORCH:
        try:
            pytorch_load(model)
            return "pytorch"
        except:
            pass

    if source == "auto" and isinstance(model, Program):
        return "mil"

    msg = (
        "Unable to determine the type of the model, i.e. the source framework. "
        'Please provide the value of argument "source", from one of '
        '["tensorflow", "pytorch", "mil"]. Note that model conversion requires the '
        "source package that generates the model. Please make sure you have "
        "the appropriate version of source package installed. E.g., if you're "
        "converting model originally trained with TensorFlow 1.14, make sure "
        "you have `tensorflow==1.14` installed."
    )
    raise ValueError(msg)


def _record_src_version(mlmodel, exact_source):
    # recording metadata: coremltools version, source framework and version
    if exact_source in {"tensorflow", "tensorflow2"} and (_HAS_TF_1 or _HAS_TF_2):
        src_pkg_version = "tensorflow=={0}".format(tf.__version__)
    elif exact_source == "pytorch" and _HAS_TORCH:
        src_pkg_version = "torch=={0}".format(torch.__version__)
    elif exact_source == 'mil':
        src_pkg_version = "mil"
    else:
        raise ValueError('Unsupported source {}'.format(exact_source))

    mlmodel.user_defined_metadata[_METADATA_SOURCE] = src_pkg_version
    return mlmodel



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import logging
import numpy as np
from coremltools.converters.mil.mil import types
from coremltools.converters.mil.mil.types import is_compatible_type
from coremltools.converters.mil.mil.types.symbolic import is_symbolic, any_symbolic
from . import SPACES
from .block import curr_block
from .input_type import TupleInputType, DefaultInputs
from .var import Var, InternalVar, ListVar

VALUE = 1
SYMBOL = 2
NONE = 4
ALL = 7


def _is_compatible_symbolic_array(a, b):
    """
    A helper function which check if two numpy array with symbolic value.
    For instance, a = np.array([is0, is2])
                  b = np.array([is1, 1])
    are considered compatible.
                  a = np.array([is0, 1])
                  b = np.array([is1, -1])
    are not.
    """
    if not a.shape == b.shape:
        return False
    a = a.flatten()
    b = b.flatten()
    for t, v in zip(a, b):
        if not is_symbolic(t) and not is_symbolic(v):
            if t != v:
                return False
    return True


def precondition(allow=ALL):
    """
    A helper decorator for value_inference method.
    Decorate value_inference with parameter VALUE/SYMBOL/NONE or ALL.
    For VALUE/SYMBOL/NONE use logical or ( | ) for multiple allowance.
    Note that:
        1. ALL == VALUE | SYMBOL | NONE
        2. Chosen flag (some or all VALUE/SYMBOL/NONE) must be satisfied
           by EVERY INPUTS for the precondition to be satisfied.

    The meaning for each flag is:
    VALUE: value that can be materialized during compile time
    SYMBOL: value that cannot be materialized by exist as a symbol value
    NONE: a None value

    Usage:
    @precondition(allow=VALUE|SYMBOL)
    def value_inference(self):
        '''some value_inference implementation'''
    """
    ALLOW_VALUE = allow & VALUE
    ALLOW_SYMBOL = allow & SYMBOL
    ALLOW_NONE = allow & NONE

    def process(v, has_value, has_symbol, has_none):
        """
        v: Var

        Return updated has_value, has_symbol, has_none
        """
        if any_symbolic(v.sym_val):
            return has_value, True, has_none
        elif v.val is None:
            return has_value, has_symbol, True
        return True, has_symbol, has_none

    def decorator(func):
        def wrapper(self):
            HAS_VALUE = False
            HAS_SYMBOL = False
            HAS_NONE = False
            for in_name, in_type in self._input_types.items():
                if in_type.optional:
                    # Optional inputs are not required to invoke value_inference()
                    continue

                if isinstance(in_type, TupleInputType):
                    for v in self._input_vars[in_name]:
                        HAS_VALUE, HAS_SYMBOL, HAS_NONE = process(
                            v, HAS_VALUE, HAS_SYMBOL, HAS_NONE
                        )
                else:
                    HAS_VALUE, HAS_SYMBOL, HAS_NONE = process(
                        self._input_vars[in_name], HAS_VALUE, HAS_SYMBOL, HAS_NONE
                    )

            if HAS_VALUE and not ALLOW_VALUE:
                msg = "Implementation of value_inference() for op {} doesn't support input with VALUE"
                raise NotImplementedError(msg.format(self.op_type))
            elif HAS_SYMBOL and not ALLOW_SYMBOL:
                msg = "Implementation of value_inference() for op {} doesn't support input with SYMBOL"
                raise NotImplementedError(msg.format(self.op_type))
            elif HAS_NONE and not ALLOW_NONE:
                msg = "Implementation of value_inference() for op {} doesn't support input with NONE"
                raise NotImplementedError(msg.format(self.op_type))
            else:
                return func(self)

        return wrapper

    return decorator


def is_internal_input(arg_name):
    return arg_name[0] == "_"


class mil_list(object):
    '''
    A wrapper around python list
    '''

    def __init__(self, ls=None):
        self.ls = ls if ls is not None else []
        if not isinstance(self.ls, list):
            raise TypeError("Type of 'ls' must be list in the 'mil_list' class")


class Operation(object):
    """
    Represents Operation in MIL.

    # Properties
    name (str):
        The name of the operation

    input_types (InputSpec, class attr):
        Read-only named input types from all subclasses. Input types are used
        to validate `inputs`.

    inputs [_input_vars] (dict of str --> Var):
        An Operation (subclass of Operation) only has access to input Var,
        which is already validated against `input_spec`.

    outputs [_output_vars] (list of Var):
        List of output var based on type inference. Read-only
    """

    def __init__(self, **kwargs):
        self._input_types = self.input_spec.input_types
        self.name = kwargs.get("name", None)

        self._output_vars = None
        self._input_vars = {}
        self.blocks = []
        self.enclosing_block = curr_block()

        # Initialize inputs as object attributes (all None)
        for k in self._input_types.keys():
            setattr(self, k, None)
            self._input_vars[k] = None

        self._check_expected_inputs(kwargs)

        # Set inputs from kwargs
        input_kv = {k: v for k, v in kwargs.items() \
            if k in self._input_types and v is not None}
        self._validate_and_set_inputs(input_kv)
        self._ensure_required_inputs()

    def _check_expected_inputs(self, kwargs):
        """
        Check that all kwargs inputs are one of the followings:

        - system inputs (non-attributes)
        - op inputs (self._input_types.keys())
        """
        non_attributes = [
            "name",
            "symbolic_datatype",
            "datatype",
            "symbolic_value",
            "value",
            "version",
            "before_op",
            "no_check_var_visibility",  # no_check_var_visibility==True to deviate from SSA
            "no_check_var_types",  # no_check_var_types==True to force set inputs, even if type does not match with earlier ones
        ]
        for k in kwargs.keys():
            if k not in non_attributes and k not in self._input_types:
                raise ValueError(
                    "Unknown input '{}' for op '{}'".format(
                      k, self.op_type)
                    )

    def set_inputs(self,
        no_check_var_types=False,
        type_inference=False,
        **input_kvs):
        """
        Parameters
        ----------
        - input_kvs: Dict[str, Var]
          Value cannot be None

        - type_inference: bool
          True to perform type inference and recreate output Var.
        """
        self._validate_and_set_inputs(input_kvs,
            no_check_var_types=no_check_var_types)
        if type_inference and not no_check_var_types:
            self.type_value_inference()
        self._ensure_required_inputs()

    def get_flattened_inputs(self):
        """
        Returns:
        list[Var]. Flatten all tuple inputs
        """
        flat_inputs = []
        for v in self.inputs.values():
            if isinstance(v, (list, tuple)):
                flat_inputs.extend(v)
            else:
                flat_inputs.append(v)
        return flat_inputs

    def type_value_inference(self, overwrite_output=False):
        """
        Perform type inference and auto_val computation based on new input Vars
        in kwargs. If self._output_vars is None then we generate _output_vars;
        otherwise no new Var is created, but type inference result is verified
        against existing _output_vars, if overwrite_output is False.

        If overwrite_output is True, then the type inference result overwrites the
        existing _output_vars
        """
        output_types = self.type_inference()
        if not isinstance(output_types, tuple):
            output_types = (output_types,)
        output_vals = self._auto_val(output_types)
        try:
            output_names = self.output_names()
            if not isinstance(output_names, tuple):
                output_names = (output_names,)
        except NotImplementedError as e:
            if len(output_types) > 1:
                output_names = tuple(str(i) for i, _ in enumerate(output_types))
            else:
                output_names = ("",)  # output name same as op name.

        # Combine (output_names, output_types, output_vals) to create output
        # Vars.
        if self._output_vars is None:
            self._output_vars = []
            for i, (n, sym_type, sym_val) in enumerate(
                zip(output_names, output_types, output_vals)
            ):
                name = self.name + ":" + n if n != "" else self.name
                if types.is_list(sym_type):
                    new_var = ListVar(
                        name,
                        elem_type=sym_type.T[0],
                        init_length=sym_type.T[1],
                        dynamic_length=sym_type.T[2],
                        sym_val=sym_val if (sym_val is not None and isinstance(sym_val.val, list)) else None,
                        op=self,
                        op_output_idx=i,
                    )
                else:
                    new_var = Var(name, sym_type, sym_val, op=self, op_output_idx=i)
                self._output_vars.append(new_var)
        else:
            # Check new inference result against existing self._output_vars.
            for i, (n, sym_type, sym_val) in enumerate(
                zip(output_names, output_types, output_vals)
            ):
                out_var = self._output_vars[i]
                # Check type inference
                if overwrite_output:
                    out_var._sym_type = sym_type
                elif not is_compatible_type(sym_type, out_var.sym_type):
                    msg = "Output Var {} in op {} type changes with new input Vars"
                    raise ValueError(msg.format(out_var.name, self.name))

                # Check value inference
                if overwrite_output:
                    out_var._sym_val = sym_val

                if sym_val is not None and out_var.sym_val is not None:
                    if np.any(sym_val.val != out_var.sym_val):
                        if overwrite_output:
                            out_var._sym_val = sym_val
                        else:
                            msg = 'value_inference differs for var {} in op {}'
                            if not _is_compatible_symbolic_array(sym_val.val, out_var.sym_val):
                                raise ValueError(msg.format(out_var.name, self.name))

    def _auto_val(self, output_types):
        """
        # Evaluation is two stage:
        #
        # Stage 1: Check whether the method value_inference() is implemented
        #
        # Stage 2: Check if there's an value_inference() implementation
        #          for given input types.
        #
        # Suppose input are all SYMBOL:
        # Case 1: No value_inference() implemented => fail at stage 1
        # Case 2: If value_inference() implemented, but requires all VALUE not
        #         SYMBOL => fail at stage 2
        # Case 3: If value_inference() implemented, and has no restriction on
        #         input types => Success
        #
        # If either stage fails, outputs[i].val is None.
        # Otherwise, output[i].sym_val is not None.

        output_types: tuple of builtin types

        Returns:
            output_vals: tuple of builtin type with value, or tuple of None
        """
        do_auto_val = True

        if do_auto_val:
            # Is self.value_inference implemented for corresponding input?
            try:
                vals = self.value_inference()
            except NotImplementedError as e:
                do_auto_val = False

        if not do_auto_val:
            # No auto_val possible.
            return tuple(None for _ in output_types)

        if not isinstance(vals, (tuple, list)):
            vals = (vals,)
        for val in vals:
            if val is None:
                do_auto_val = False
        if not do_auto_val:
            # No auto_val possible.
            return tuple(None for _ in output_types)

        auto_val = []
        for t, v in zip(output_types, vals):
            builtin_val = t()
            if isinstance(v, mil_list):
                builtin_val.val = v.ls
            else:
                builtin_val.val = v
            auto_val.append(builtin_val)
        return auto_val

    def value_inference(self):
        """
        Optional Python implementation of the op based on (materialized) values
        in `self.input_var`. Return a builtin value (single output) or a tuple of
        builtin values (multi-outputs) of the same length as returned by `
        type_inference`
        """
        msg = "value_inference() is not implemented by op {}"
        raise NotImplementedError(msg.format(self.op_type))

    def default_inputs(self):
        """
        Optional. Returns default values for optional inputs. The
        function is guaranteed to have access to all required inputs and
        possibly some optional inputs should the user supply them.
        They may be used to construct default values, such as
        `strides=[1]*num_spatial_dims` in conv, where
        `num_spatial_dims` may be inferred from the rank of
        required inputs
        """
        return DefaultInputs()

    def output_names(self):
        """
        Optional. If implemented, we set the output var i name as
        self.name + "/" + output_names[i]

        Returns a string (single output) or tuple of strings
        """
        msg = "output_names() is not implemented by op {}"
        raise NotImplementedError(msg.format(self.op_type))

    def type_inference(self):
        """
        Return (builtin_type, builtin_val) pair from type inference.
        builtin_val may be None if symbolic_value is not attainable at compile
        time.
        """
        raise NotImplementedError("This function must be implemented by each op")

    def build_nested_blocks(self):
        """
        Build nested blocks (for cond and while_loop and other composite
        blocks)
        """
        pass

    def _ensure_required_inputs(self):
        """
        Raise value error if required inputs aren't present
        """
        for name, input_type in self._input_types.items():
            if not input_type.optional and \
                self._input_vars[name] is None:
                msg_prefix = 'Op \"{}\" (op_type: {}) '.format(self.name,
                self.op_type)
                raise ValueError(msg_prefix + \
                    "Required input {} is missing".format(name))

    def _validate_and_set_inputs(self, input_kvs,
        no_check_var_types=False):
        """
        For each k, v in `input_kvs`, perform the followings:

        - Check k exists in `self.input_specs`
        - Check that v satisfies the correspodning `InputType`
        - Set input, possibly replacing existing input.

        Note that it does not ensure all required inputs are satisfied.
        Use _ensure_required_inputs() for that.

        Parameters
        ----------
        - input_kvs: Dict[str, Var]
          Each key in input_kvs must exist in `self.input_specs`. Its values
          must be a Var.

        - no_check_var_types: bool
          True to check var types against input_specs only, but not
          enforcing new input vars to be a subtype of existing input vars
        """
        for key in input_kvs.keys():
            if key not in self._input_types:
                raise RuntimeError(
                    "Unknown input '{}' for op '{}'".format(key, self.op_type)
                )

        def check_and_detach(v_new, v_old, op, no_check_var_types):
            # Check new var's sym_type is compatible with the
            # existing's sym_type.
            if (
                not is_compatible_type(v_new.sym_type, v_old.sym_type)
                and not no_check_var_types
            ):
                msg = "New var type {} not a subtype of " + "existing var type {}"
                raise ValueError(msg.format(v_new.sym_type, v_old.sym_type))
            v_old.remove_child_op(op, no_check_var_types)

        self.input_spec.validate_inputs(self.name, self.op_type, input_kvs)

        for name, var in input_kvs.items():
            # TODO: remove InternalVar check
            #if not isinstance(var, InternalVar):

            # Remove this operation itself from existing input
            # Var's child_ops
            existing_input_var = self._input_vars[name]
            if existing_input_var is not None:
                if isinstance(existing_input_var, (list, tuple)):
                    for v_old, v_new in zip(existing_input_var, var):
                        check_and_detach(v_new, v_old, self, no_check_var_types)
                else:
                    check_and_detach(
                        var, existing_input_var, self, no_check_var_types
                    )

            # Set var as input_var
            if isinstance(var, Var):
                var.add_child_op(self)
            elif isinstance(var, (tuple, list)):
                for v in var:
                    v.add_child_op(self)
            # ignore function inputs
            self._input_vars[name] = var
            setattr(self, name, var)


    @property
    def inputs(self):
        """
        Returns
        -------
        - inputs: Dict[str, Union[Var, Tuple[Var]]]
        """
        # Filter out InternalVar
        return {k: v for k, v in self._input_vars.items() if not
            isinstance(v, InternalVar) and v is not None}

    @property
    def outputs(self):
        return self._output_vars

    @property
    def op_type(self):
        return type(self).__name__

    def remove_from_block(self):
        """
        Remove / detach itself from the enclosing block. See Block.remove_ops
        for details.
        """
        self.enclosing_block.remove_ops([self])

    @staticmethod
    def var_to_str(v):
        if isinstance(v, (tuple, list)):
            return "(" + ", ".join(["%" + s.name for s in v]) + ")"
        else:
            return "%" + v.name

    def indented_str(self, indent=""):
        s = indent
        if self.outputs is not None:
            s += ", ".join([str(o) for o in self.outputs])
        s += " = " + self.op_type + "("
        if self.op_type == "const":
            if self.mode.val == "immediate_value":
                if isinstance(self.val.sym_val, (np.generic, np.ndarray)):
                    val_str = str(self.val.sym_val.tolist())
                else:
                    val_str = (
                        '"' + self.val.sym_val + '"'
                        if isinstance(self.val.sym_val, str)
                        else str(self.val.sym_val)
                    )
                s += "val=" + val_str
            else:
                s += "val=(file_value)"
        else:
            s += ", ".join(
                [
                    k + "=" + Operation.var_to_str(self.inputs[k])
                    for k in self._input_types.keys()
                    if k in self.inputs and not is_internal_input(k)
                ]
            )
        s += ', name="{}")\n'.format(self.name)
        for b in self.blocks:
            s += b.indented_str(indent=indent + SPACES)
        return s

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.indented_str(SPACES)
